from abc import ABCMeta, abstractmethod

from deprecation import deprecated
from meiga import (
    BoolResult,
    Error,
    Failure,
    NotImplementedMethodError,
    Result,
    Success,
    isSuccess,
)

from petisco import Uuid
from petisco.base.application.patterns.repository import Repository
from petisco.extra.sqlalchemy.sql.errors import (
    ClientNotFoundError,
    EntitiesNotFoundError,
    EntityAlreadyExistError,
    EntityNotFoundError,
    UserNotFoundError,
)


class SqlRepository(Repository, metaclass=ABCMeta):
    def get_sql_internal_client_id(self, session, model) -> Result[int, Error]:
        if not hasattr(self, "internal_client_id"):
            client_id = self.get_info_id().client_id
            internal_client_id = (
                session.query(model.id)
                .filter(model.client_id == client_id.value)
                .first()
            )
            if not internal_client_id:
                return Failure(ClientNotFoundError(client_id))
            self.internal_client_id = internal_client_id.id

        return Success(self.internal_client_id)

    def get_sql_internal_user_id(self, session, model) -> Result[int, Error]:
        if not hasattr(self, "internal_user_id"):
            user_id = self.get_info_id().user_id
            internal_user_id = (
                session.query(model.id).filter(model.user_id == user_id.value).first()
            )
            if not internal_user_id:
                return Failure(UserNotFoundError(user_id))
            self.internal_user_id = internal_user_id.id

        return Success(self.internal_user_id)

    @classmethod
    def fail_if_entity_already_exist(
        cls, model, entity_id: Uuid, result_error: Error = None
    ) -> BoolResult:
        if model:
            error = (
                EntityAlreadyExistError(cls.__name__, model.__tablename__, entity_id)
                if not result_error
                else result_error
            )
            return Failure(error)
        return isSuccess

    @classmethod
    def fail_if_entity_not_found(
        cls, model, entity_id: Uuid, result_error: Error = None
    ) -> BoolResult:
        if not model:
            error = (
                EntityNotFoundError(cls.__name__, entity_id)
                if not result_error
                else result_error
            )
            return Failure(error)
        return isSuccess

    @classmethod
    def fail_if_entities_not_found(
        cls, model, result_error: Error = None
    ) -> BoolResult:
        if not model:
            error = (
                EntitiesNotFoundError(cls.__name__)
                if not result_error
                else result_error
            )
            return Failure(error)
        return isSuccess

    @abstractmethod
    def save(self, *args, **kwargs) -> BoolResult:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve_all(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def remove(self, *args, **kwargs) -> BoolResult:
        return NotImplementedMethodError

    @classmethod
    @deprecated("This method is deprecated. Please, use fail_if_entity_not_found")
    def fail_if_entity_not_exist(
        cls, model, entity_id: Uuid, result_error: Error = None
    ) -> BoolResult:
        return cls.fail_if_entity_not_found(model, entity_id, result_error)

    @classmethod
    @deprecated("This method is deprecated. Please, use fail_if_entities_not_found")
    def fail_if_entities_not_exist(
        cls, model, result_error: Error = None
    ) -> BoolResult:
        return cls.fail_if_entities_not_found(model, result_error)



from typing import Dict

from petisco.legacy import IAppService, AppServices
from tests.end2end.flask_app.sum.infrastructure.sum_executor import SumExecutor


def load_app_services():
    AppServices.load(app_services_provider)


def app_services_provider() -> Dict[str, IAppService]:
    return {"sum": SumExecutor()}



from abc import ABCMeta
from typing import Dict

from petisco.legacy.domain.aggregate_roots.info_id import InfoId
from petisco.legacy.domain.value_objects.client_id import ClientId
import copy

from petisco.legacy.domain.value_objects.user_id import UserId


class PatternBase:

    __metaclass__ = ABCMeta

    @classmethod
    def info(cls) -> Dict:
        return {"name": cls.__name__}

    def _set_client_id(self, client_id: ClientId):
        if client_id is not None:
            self.client_id = client_id

    def _set_user_id(self, user_id: UserId):
        if user_id is not None:
            self.user_id = user_id

    def _set_info_id(self, info_id: InfoId):
        self._set_client_id(info_id.client_id)
        self._set_user_id(info_id.user_id)
        self.info_id = info_id

    def with_client_id(self, client_id: ClientId):
        repository = copy.copy(self)
        repository._set_client_id(client_id)
        return repository

    def with_info_id(self, info_id: InfoId):
        repository = copy.copy(self)
        repository._set_info_id(info_id)
        return repository

    def get_client_id(self) -> ClientId:
        if not hasattr(self, "client_id"):
            name = self.info().get("name")
            raise AttributeError(
                f"{name} needs client_id. Please, use {name}.with_client_id() to get a valid object with a client_id"
            )
        return self.client_id

    def get_client_id_value(self) -> str:
        return self.get_client_id().value

    def get_user_id(self) -> UserId:
        if not hasattr(self, "user_id"):
            name = self.info().get("name")
            raise AttributeError(
                f"{name} needs user_id. Please, use {name}.with_info_id() to get a valid with a info_id (client_id and user_id)"
            )
        return self.user_id

    def get_user_id_value(self) -> str:
        return self.get_user_id().value

    def get_info_id(self) -> InfoId:
        if not hasattr(self, "info_id"):
            name = self.info().get("name")
            raise AttributeError(
                f"{name} needs info_id. Please, use {name}.with_info_id() to get a valid with a info_id"
            )
        return self.info_id



import os

import pytest

from petisco.legacy import Event, EventId, NotifierMessage, InfoId, Petisco
from petisco.legacy.domain.value_objects.user_id import UserId
from petisco.legacy.notifier.domain.notifier_exception_message import (
    NotifierExceptionMessage,
)


class TrackedEventsSpy:
    def __init__(self):
        self.events = {}

    def append(self, event: Event):
        if str(event.event_id) not in self.events:
            self.events[str(event.event_id)] = {"counter": 1}
        else:
            self.events[str(event.event_id)]["counter"] += 1

    def assert_number_events(self, expected_number_events: int):
        actual_number_events = len(self.events.keys())
        assert (
            actual_number_events == expected_number_events
        ), f"Expected events is {expected_number_events}, actual {actual_number_events}. [{self.events}]"

    def get_counter_by_event_id(self, event_id: EventId):
        return self.events.get(str(event_id), {}).get("counter", 0)


@pytest.fixture
def make_user_created_event(given_any_user_id):
    class UserCreated(Event):
        user_id: UserId

        def __init__(self, user_id: UserId):
            self.user_id = user_id
            super().__init__()

    def _make_any_user_created_event(user_id=given_any_user_id):
        return UserCreated(user_id=user_id)

    return _make_any_user_created_event


@pytest.fixture
def make_first_name_added_event(given_any_user_id, given_any_name):
    class FirstNameAdded(Event):
        user_id: UserId
        first_name: str

        def __init__(self, user_id: UserId, first_name: str):
            self.user_id = user_id
            self.first_name = first_name
            super().__init__()

    def _make_any_first_name_added_event(
        user_id=given_any_user_id, first_name=given_any_name
    ):
        return FirstNameAdded(user_id=user_id, first_name=first_name)

    return _make_any_first_name_added_event


@pytest.fixture
def given_any_message():
    return "Hello world"


@pytest.fixture
def given_any_title():
    return "This is a title"


@pytest.fixture
def given_any_info_petisco():
    return {
        "app_name": "test",
        "app_version": "0.0.1",
        "petisco_version": "0.0.1",
        "environment": "Production",
    }


@pytest.fixture
def given_any_basic_notifier_message(given_any_message):
    return NotifierMessage(message=given_any_message)


@pytest.fixture
def given_any_notifier_message_with_info_id(
    given_any_client_id, given_any_user_id, given_any_message
):
    return NotifierMessage(
        message=given_any_message,
        info_id=InfoId(client_id=given_any_client_id, user_id=given_any_user_id),
    )


@pytest.fixture
def given_any_complete_notifier_message(
    given_any_client_id,
    given_any_user_id,
    given_any_message,
    given_any_info_petisco,
    given_any_title,
):
    return NotifierMessage(
        title=given_any_title,
        message=given_any_message,
        info_id=InfoId(client_id=given_any_client_id, user_id=given_any_user_id),
        info_petisco=given_any_info_petisco,
    )


@pytest.fixture
def given_any_exception():
    return ValueError("Oops, something went wrong")


@pytest.fixture
def given_any_traceback():
    return """Traceback (most recent call last):
  File "e.py", line 7, in <module>
    raise TypeError("Again !?!")
TypeError: Again !?!
"""


@pytest.fixture
def given_long_traceback():
    return (
        """Traceback (most recent call last):
  File "e.py", line 7, in <module>
    raise TypeError("Again !?!")
TypeError: Again !?!
"""
        * 500
    )


@pytest.fixture
def given_any_input_parameters():
    return {"parameter_1": 1, "parameter_2": 2}


@pytest.fixture
def given_any_complete_notifier_exception_message(
    given_any_client_id,
    given_any_user_id,
    given_any_message,
    given_any_info_petisco,
    given_any_exception,
    given_any_input_parameters,
    given_any_traceback,
):
    return NotifierExceptionMessage(
        executor="test_function",
        exception=given_any_exception,
        input_parameters=given_any_input_parameters,
        traceback=given_any_traceback,
        info_id=InfoId(client_id=given_any_client_id, user_id=given_any_user_id),
        info_petisco=given_any_info_petisco,
    )


@pytest.fixture
def given_long_traceback_complete_notifier_exception_message(
    given_any_client_id,
    given_any_user_id,
    given_any_message,
    given_any_info_petisco,
    given_any_exception,
    given_long_traceback,
):
    return NotifierExceptionMessage(
        executor="test_function",
        exception=given_any_exception,
        traceback=given_long_traceback,
        info_id=InfoId(client_id=given_any_client_id, user_id=given_any_user_id),
        info_petisco=given_any_info_petisco,
    )


@pytest.fixture
def initialized_petisco():
    filename = (
        f"{os.path.dirname(os.path.abspath(__file__))}/end2end/flask_app/petisco.yml"
    )
    Petisco.clear()
    Petisco.from_filename(filename)

    yield

    Petisco.clear()



import threading
import traceback
from dataclasses import dataclass
from time import sleep
from typing import Callable, List, Optional, Type

from meiga import Failure
from pika import BasicProperties
from pika.adapters.blocking_connection import BlockingChannel
from pika.spec import Basic

from petisco.base.domain.message.chaos.message_chaos import MessageChaos
from petisco.base.domain.message.chaos.message_chaos_error import MessageChaosError
from petisco.base.domain.message.chaos.not_implemented_message_chaos import (
    NotImplementedMessageChaos,
)
from petisco.base.domain.message.command import Command
from petisco.base.domain.message.consumer_derived_action import ConsumerDerivedAction
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_handler_returns_none_error import (
    MessageHandlerReturnsNoneError,
)
from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.extra.rabbitmq.application.message.bus.rabbitmq_command_bus import (
    RabbitMqCommandBus,
)
from petisco.extra.rabbitmq.application.message.bus.rabbitmq_domain_event_bus import (
    RabbitMqDomainEventBus,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_event_consumer_logger import (
    RabbitMqMessageConsumerLogger,
)
from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_subscriber_queue_name_formatter import (
    RabbitMqMessageSubscriberQueueNameFormatter,
)
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector
from petisco.extra.rabbitmq.shared.rabbitmq_exchange_name_formatter import (
    RabbitMqExchangeNameFormatter,
)
from petisco.legacy.event.consumer.infrastructure.rabbitmq_event_consumer_printer import (
    RabbitMqEventConsumerPrinter,
)
from petisco.legacy.event.shared.infrastructure.rabbitmq.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.legacy.logger.interface_logger import ILogger
from petisco.legacy.logger.not_implemented_logger import NotImplementedLogger


@dataclass
class SubscriberItem:
    queue_name: str
    subscriber: MessageSubscriber
    consumer_tag: str
    is_store: bool = False


class RabbitMqMessageConsumer(MessageConsumer):
    def __init__(
        self,
        organization: str,
        service: str,
        max_retries: int,
        connector: RabbitMqConnector = RabbitMqConnector(),
        verbose: bool = False,
        chaos: MessageChaos = NotImplementedMessageChaos(),
        logger: Optional[ILogger] = NotImplementedLogger(),
    ):
        self.connector = connector
        self.organization = organization
        self.service = service
        self.exchange_name = f"{organization}.{service}"
        self.rabbitmq_key = f"consumer-{self.exchange_name}"
        self._fallback_store_exchange_name = f"retry.{organization}.store"
        self.max_retries = max_retries
        self._channel = self.connector.get_channel(self.rabbitmq_key)
        self.printer = RabbitMqEventConsumerPrinter(verbose)
        self.consumer_logger = RabbitMqMessageConsumerLogger(logger)
        self.chaos = chaos
        self.subscribers = {}

    def start(self):
        if not self._channel:
            raise RuntimeError(
                "RabbitMqMessageConsumer: cannot start consuming event without any subscriber defined."
            )
        self._thread = threading.Thread(target=self._start)
        self._thread.start()

    def _start(self):
        self._channel.start_consuming()

    def add_subscribers(self, subscribers: List[Type[MessageSubscriber]]):
        for SubscriberClass in subscribers:

            subscriber: MessageSubscriber = SubscriberClass()

            for subscriber_info in subscriber.get_message_subscribers_info():

                if subscriber_info.message_type == "message":
                    self._add_subscriber_instance_on_queue(
                        queue_name="store", subscriber=subscriber, is_store=True
                    )
                else:
                    queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format(
                        subscriber_info, exchange_name=self.exchange_name
                    )
                    self._add_subscriber_instance_on_queue(
                        queue_name=f"{queue_name}.{subscriber.get_subscriber_name()}",
                        subscriber=subscriber,
                        message_type_expected=subscriber_info.message_type,
                    )

    def add_subscriber_on_dead_letter(self, subscriber: Type[MessageSubscriber]):
        subscriber: MessageSubscriber = subscriber()
        for subscriber_info in subscriber.get_message_subscribers_info():

            queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format_dead_letter(
                subscriber_info, exchange_name=self.exchange_name
            )
            self._add_subscriber_instance_on_queue(
                queue_name=f"{queue_name}.{subscriber.get_subscriber_name()}",
                subscriber=subscriber,
                message_type_expected=subscriber_info.message_type,
            )

    # def add_handler_on_store(self, handler: Callable):
    #     self.add_handler_on_queue("store", handler, is_store=True)

    def _add_subscriber_instance_on_queue(
        self,
        queue_name: str,
        subscriber: MessageSubscriber,
        is_store: bool = False,
        message_type_expected: str = "message",
    ):
        consumer_tag = self._channel.basic_consume(
            queue=queue_name,
            on_message_callback=self.consumer(
                subscriber, is_store, message_type_expected
            ),
        )
        self.subscribers[queue_name] = SubscriberItem(
            queue_name, subscriber, consumer_tag, is_store
        )

    def add_subscriber_on_queue(
        self,
        queue_name: str,
        subscriber: Type[MessageSubscriber],
        is_store: bool = False,
        message_type_expected: str = "message",
    ):
        subscriber: MessageSubscriber = subscriber()
        return self._add_subscriber_instance_on_queue(
            queue_name=queue_name,
            subscriber=subscriber,
            is_store=is_store,
            message_type_expected=message_type_expected,
        )

    def consumer(
        self,
        subscriber: MessageSubscriber,
        is_store: bool = False,
        message_type_expected: str = "message",
    ) -> Callable:
        def rabbitmq_consumer(
            ch: BlockingChannel,
            method: Basic.Deliver,
            properties: BasicProperties,
            body: bytes,
        ):
            self.printer.print_received_message(method, properties, body)

            if self.chaos.nack_simulation(ch, method):
                self.consumer_logger.log_nack_simulation(
                    method, properties, body, subscriber.handle
                )
                return
            else:
                self.consumer_logger.log(
                    method,
                    properties,
                    body,
                    subscriber.handle,
                    log_activity="received_message",
                )

            try:
                if message_type_expected == "domain_event":
                    message = DomainEvent.from_json(body)
                elif message_type_expected == "command":
                    message = Command.from_json(body)
                else:
                    message = Message.from_json(body)
            except Exception as e:
                self.consumer_logger.log_parser_error(
                    method, properties, body, subscriber.handle, e
                )
                ch.basic_nack(delivery_tag=method.delivery_tag)
                return

            self.chaos.delay()

            if self.chaos.failure_simulation(method):
                self.consumer_logger.log_failure_simulation(
                    method, properties, body, subscriber.handle
                )
                result = Failure(MessageChaosError())
            else:
                connector = RabbitMqConsumerConnector(ch)
                domain_event_bus = RabbitMqDomainEventBus(
                    self.organization, self.service, connector
                )
                command_bus = RabbitMqCommandBus(
                    self.organization, self.service, connector
                )
                subscriber.set_domain_event_bus(domain_event_bus)
                subscriber.set_command_bus(command_bus)

                result = subscriber.handle(message)

            self.printer.print_context(subscriber.handle, result)

            if result is None:
                raise MessageHandlerReturnsNoneError(subscriber.handle)

            derived_action = ConsumerDerivedAction()
            if result.is_failure:
                if not properties.headers:
                    properties.headers = {
                        "queue": f"{method.routing_key}.{subscriber.get_subscriber_name()}"
                    }
                derived_action = self.handle_consumption_error(
                    ch, method, properties, body, is_store
                )
            else:
                ch.basic_ack(delivery_tag=method.delivery_tag)

            self.consumer_logger.log(
                method,
                properties,
                body,
                subscriber.handle,
                "computed_message",
                result,
                derived_action,
            )
            self.printer.print_separator()

        return rabbitmq_consumer

    def handle_consumption_error(
        self,
        ch: BlockingChannel,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        is_store: bool,
    ):

        if self.has_been_redelivered_too_much(properties):
            derived_action = self.send_to_dead_letter(ch, method, properties, body)
        else:
            derived_action = self.send_to_retry(ch, method, properties, body, is_store)

        ch.basic_ack(delivery_tag=method.delivery_tag)

        return derived_action

    def has_been_redelivered_too_much(self, properties: BasicProperties):
        if not properties.headers or "redelivery_count" not in properties.headers:
            if self.max_retries < 1:
                return True
            return False
        else:
            return properties.headers.get("redelivery_count") >= self.max_retries

    def _get_routing_key(self, routing_key: str, prefix: str):
        if routing_key.startswith("retry."):
            routing_key = routing_key.replace("retry.", prefix, 1)
        elif routing_key.startswith("dead_letter."):
            routing_key = routing_key.replace("dead_letter.", prefix, 1)
        else:
            routing_key = f"{prefix}{routing_key}"
        return routing_key

    def send_to_retry(
        self,
        ch: BlockingChannel,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        is_store: bool = False,
    ):
        self.printer.print_action("send_to_retry")
        exchange_name = RabbitMqExchangeNameFormatter.retry(self.exchange_name)

        routing_key = method.routing_key
        if properties.headers:
            routing_key = properties.headers.get("queue", routing_key)

        if is_store:
            routing_key = "store"
            exchange_name = self._fallback_store_exchange_name

        routing_key = self._get_routing_key(routing_key, "retry.")

        updated_headers = self.send_message_to(
            exchange_name, ch, routing_key, properties, body
        )
        return ConsumerDerivedAction(
            action="send_to_retry",
            exchange_name=exchange_name,
            routing_key=routing_key,
            headers=updated_headers,
        )

    def send_to_dead_letter(
        self,
        ch: BlockingChannel,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
    ):
        self.printer.print_action("send_to_dead_letter")
        exchange_name = RabbitMqExchangeNameFormatter.dead_letter(self.exchange_name)
        routing_key = self._get_routing_key(method.routing_key, "dead_letter.")
        updated_headers = self.send_message_to(
            exchange_name, ch, routing_key, properties, body
        )
        return ConsumerDerivedAction(
            action="send_to_dead_letter",
            exchange_name=exchange_name,
            routing_key=routing_key,
            headers=updated_headers,
        )

    def send_message_to(
        self,
        exchange_name: str,
        ch: BlockingChannel,
        routing_key: str,
        properties: BasicProperties,
        body: bytes,
    ):
        if properties.headers:
            redelivery_count = properties.headers.get("redelivery_count", 0)
            properties.headers["redelivery_count"] = redelivery_count + 1
        else:
            properties.headers = {"redelivery_count": 1}

        self.printer.print_send_message_to(
            exchange_name, routing_key, properties.headers
        )

        ch.basic_publish(
            exchange=exchange_name,
            routing_key=routing_key,
            body=body,
            properties=properties,
        )

        return properties.headers

    def stop(self):
        def _log_stop_exception(e: Exception):
            from petisco.legacy import ERROR, LogMessage, Petisco

            logger = Petisco.get_logger()
            log_message = LogMessage(
                layer="petisco", operation="RabbitMQEventSubscriber"
            )
            message = f"Error stopping RabbitMQEventSubscriber: {repr(e.__class__)} {e} | {traceback.format_exc()}"
            logger.log(ERROR, log_message.set_message(message))

        if self._thread and self._thread.is_alive():
            self._unsubscribe_all()
            try:
                self._thread.join()
                self._thread = None
            except Exception as e:
                _log_stop_exception(e)

    def _unsubscribe_all(self):
        def _stop_consuming_consumer_channel():
            try:
                self._channel.stop_consuming()
                self._channel.cancel()
            except ValueError:
                pass

        def _await_for_stop_consuming_consumer_channel():
            sleep(2.0)

        self.connector.get_connection(self.rabbitmq_key).call_later(
            0, _stop_consuming_consumer_channel
        )

        _await_for_stop_consuming_consumer_channel()

    def unsubscribe_subscriber_on_queue(self, queue_name: str):
        subscriber_item: SubscriberItem = self.subscribers.get(queue_name)
        if subscriber_item is None:
            raise IndexError(
                f"Cannot unsubscribe an nonexistent queue ({queue_name}). Please, check configured consumers ({list(self.subscribers.keys())})"
            )

        def _unsubscribe_handler_on_queue():
            self._channel.basic_cancel(consumer_tag=subscriber_item.consumer_tag)

        self._do_it_in_consumer_thread(_unsubscribe_handler_on_queue)

    def resume_subscriber_on_queue(self, queue_name: str):
        subscriber_item: SubscriberItem = self.subscribers.get(queue_name)
        if subscriber_item is None:
            raise IndexError(
                f"Cannot resume an nonexistent queue ({queue_name}). Please, check configured consumers ({list(self.subscribers.keys())})"
            )

        def _resume_handler_on_queue():
            subscriber_item.consumer_tag = self._channel.basic_consume(
                queue=subscriber_item.queue_name,
                on_message_callback=self.consumer(
                    subscriber_item.subscriber, subscriber_item.is_store
                ),
            )

        self._do_it_in_consumer_thread(_resume_handler_on_queue)

    def _do_it_in_consumer_thread(self, action: Callable):
        def _execute_action():
            try:
                action()
            except ValueError:
                pass

        def _await_for_thread():
            sleep(2.0)

        self.connector.get_connection(self.rabbitmq_key).call_later(0, _execute_action)

        _await_for_thread()



from pymongo import MongoClient

from petisco.legacy.application.singleton import Singleton


class PyMongoPersistence(metaclass=Singleton):
    @staticmethod
    def get_instance():
        return PyMongoPersistence()

    def __init__(self, client: MongoClient = None, database: str = None):
        self.client = client
        self.database = database



from abc import ABC, abstractmethod
from functools import wraps
from types import FunctionType
from typing import Any

from meiga import Result, Error, NotImplementedMethodError, Failure
from meiga.on_failure_exception import OnFailureException

from petisco.base.application.use_case.use_case_uncontrolled_error import (
    UseCaseUncontrolledError,
)


def wrapper(method):
    @wraps(method)
    def wrapped(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except OnFailureException as exception:
            return exception.result
        except Error as error:
            return Failure(error)
        except Exception as exception:
            return Failure(UseCaseUncontrolledError(exception))

    return wrapped


class MetaUseCase(type, ABC):
    def __new__(mcs, name, bases, namespace):
        new_class_dict = {}
        if "execute" not in namespace:
            raise NotImplementedError(
                "Petisco UseCase must implement an execute method"
            )

        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                attribute = wrapper(attribute)
            new_class_dict[attributeName] = attribute
        return type.__new__(mcs, name, bases, new_class_dict)

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError


class UseCase(metaclass=MetaUseCase):
    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError



from meiga import Result, Error, Success

from petisco.legacy import UseCase, use_case_handler, InfoId, Repositories
from petisco.legacy.domain.value_objects.name import Name
from tests.end2end.flask_app.shared.domain.repositories.interface_user_repository import (
    IUserRepository,
)


@use_case_handler()
class GetUserName(UseCase):
    @staticmethod
    def build():
        return GetUserName(repository=Repositories.get("user"))

    def __init__(self, repository: IUserRepository):
        self.repository = repository

    def execute(self, info_id: InfoId) -> Result[Name, Error]:
        user = self.repository.retrieve(
            client_id=info_id.client_id, user_id=info_id.user_id
        ).unwrap_or_return()
        return Success(user.name)



from petisco.legacy.application.singleton import Singleton


class SqlAlchemyPersistence(metaclass=Singleton):
    @staticmethod
    def get_instance():
        return SqlAlchemyPersistence()

    def __init__(self):
        self.sources = {}



from typing import Optional, Dict

from dataclasses import dataclass

from petisco.legacy.tasks.config.config_task import ConfigTask


@dataclass
class ConfigTasks:
    tasks: Optional[Dict[str, ConfigTask]] = None

    @staticmethod
    def from_dict(kdict):
        if not kdict or not isinstance(kdict, dict):
            return ConfigTasks({})

        tasks = {}
        for key, config_task in kdict.items():
            tasks[key] = ConfigTask.from_dict(config_task)

        return ConfigTasks(tasks)



from typing import Any, Dict

from dataclasses import dataclass

from petisco.legacy.domain.aggregate_roots.info_id import InfoId


@dataclass
class LogMessage:
    layer: str = None
    operation: str = None
    info_id: InfoId = None
    data: Dict[str, Any] = None

    def to_dict(self):
        dict_log_message = {"data": self.data, "meta": {}}
        if self.layer:
            dict_log_message["meta"]["layer"] = self.layer
        if self.operation:
            dict_log_message["meta"]["operation"] = self.operation
        if self.info_id:
            info_id: dict = self.info_id.to_dict()
            dict_log_message["meta"]["info_id"] = info_id

        return dict_log_message

    def set_message(self, message: Any):
        self.data = {"message": message}
        return self



from abc import abstractmethod

from meiga import NotImplementedMethodError, Result

from petisco.base.application.patterns.base_pattern import BasePattern


class AppService(BasePattern):
    @abstractmethod
    def execute(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError



from meiga import Result

from petisco.legacy import controller_handler
from tests.end2end.flask_app.sum.infrastructure.sum_executor import SumExecutor


def success_handler(result: Result):
    return {"result": result.value}, 200


@controller_handler(success_handler=success_handler)
def sum(body):

    value_1 = body.get("value_1")
    value_2 = body.get("value_2")

    return SumExecutor().execute(value_1, value_2)



import inspect
import json
import traceback
from typing import Tuple, Dict

from petisco.legacy.event.shared.domain.event import Event

REQUEST_RESPONDED_UNWRAP_ERROR = (
    {
        "error": {
            "type": "RequestRespondedUnwrapError",
            "message": "Cannot unwrap http response from decorated function",
        }
    },
)


def is_flask_response(value):
    try:
        from flask import Response

        if isinstance(value, Response):
            return True
    except (RuntimeError, ImportError):
        pass
    return False


def is_success(status_code: int):
    return 200 >= status_code <= 299


def get_content(response, status_code):
    if is_flask_response(response):
        return {"message": "flask response"}
    else:
        content = response
        if isinstance(content, str):
            content = json.loads(content)
        content_size = len(str(content))

        if content_size > 300:
            if is_success(status_code):
                content = {
                    "message": "Response OK (Trimmed message: message too long)",
                    "message_size": content_size,
                }
            else:
                error_type = content.get("error", {}).get("type", "Error")
                content = {
                    "error": {
                        "type": error_type,
                        "message": "Response Error (Trimmed message: message too long)",
                    },
                    "message_size": content_size,
                }
        else:
            content = {"message": json.dumps(content), "message_size": content_size}
        return content


class RequestResponded(Event):
    app_name: str
    app_version: str
    controller: str
    is_success: bool
    http_response: Dict
    elapsed_time: float
    event_version: str = None

    def __init__(
        self,
        app_name: str,
        app_version: str,
        controller: str,
        is_success: bool,
        http_response: Tuple[Dict, int],
        elapsed_time: float,
    ):
        self.app_name = app_name
        self.app_version = app_version
        self.controller = controller
        self.is_success = is_success
        self.elapsed_time = elapsed_time
        self.set_http_response(http_response)
        super().__init__()

    def set_http_response(self, http_response):
        try:
            _http_response = {
                "content": REQUEST_RESPONDED_UNWRAP_ERROR,
                "status_code": 500,
            }
            if isinstance(http_response, Tuple):
                _http_response["content"] = get_content(
                    http_response[0], http_response[1]
                )
                _http_response["status_code"] = http_response[1]
            else:
                _http_response["content"] = get_content(
                    http_response, http_response.status_code
                )
                _http_response["status_code"] = http_response.status_code
        except Exception as e:  # noqa E722
            traceback.print_exc()
            frame_info = inspect.stack()[1]
            raise ImportError(
                f"Error parsing response on Petisco RequestResponded Event\n"
                f"\tfilename: {frame_info.filename}\n"
                f"\tlineno: {frame_info.lineno}\n"
                f"\tfunction: {frame_info.function}\n"
                f"\tcode_context: {frame_info.code_context}"
            )
        self.http_response = _http_response



from typing import List

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.misc.singleton import Singleton


class Injector(metaclass=Singleton):
    """@DynamicAttrs"""

    def __init__(self):
        self.available_dependencies = []

    @staticmethod
    def get(name: str):
        injector = Injector()
        if not hasattr(injector, name):
            raise IndexError(
                f"Invalid dependency. {name} is not found within available dependencies [{injector.get_available_dependencies()}]"
            )
        return getattr(injector, name)

    @staticmethod
    def set_dependencies(dependencies: List[Dependency] = []):
        Injector()._set_dependencies(dependencies)

    @staticmethod
    def get_available_dependencies() -> List[str]:
        return Injector().available_dependencies

    def _set_dependencies(self, dependencies: List[Dependency] = []) -> List[str]:
        for dependency in dependencies:
            setattr(self, dependency.name, dependency.get_instance())
            if dependency.name not in self.available_dependencies:
                self.available_dependencies.append(dependency.name)



from typing import Callable, Dict

from dataclasses import dataclass

from petisco.legacy.application.config.get_funtion_from_string import (
    get_function_from_string,
)
from petisco.legacy.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


def get_float_if_exist(kdict: Dict, value: str, default_value=None) -> float:
    value = kdict.get(value)
    if value:
        value = float(value)
    else:
        value = default_value
    return value


def get_type(run_in, interval, every) -> str:
    if interval:
        type_task = "recurring"
    else:

        if run_in:
            type_task = "scheduled"
        else:
            if every:
                type_task = "cron"
            else:
                type_task = "instant"
    return type_task


@dataclass
class ConfigTask:
    handler_key: str
    run_in: float = 0.0
    interval: float = None
    type: str = "instant"
    every: str = None
    kdict: Dict = None

    def to_dict(self):
        kdict = {"type": self.type}
        if self.run_in is not None:
            kdict["run_in"] = self.run_in
        if self.interval is not None:
            kdict["interval"] = self.interval
        if self.every is not None:
            kdict["every"] = self.every
        return kdict

    @staticmethod
    def from_dict(kdict):
        run_in = get_float_if_exist(kdict, "run_in", default_value=0.0)
        interval = get_float_if_exist(kdict, "interval")
        handler_key = kdict.get("handler")
        every = kdict.get("every", None)
        type = get_type(run_in, interval, every)

        if not handler_key:
            raise TypeError("ConfigTask: handler is a required parameter")

        return ConfigTask(
            run_in=run_in,
            handler_key=handler_key,
            interval=interval,
            every=every,
            type=type,
            kdict=kdict,
        )

    def get_handler(self) -> Callable:
        # Use this function to avoid circular dependency accessing Petisco objects before its configuration
        handler = (
            get_function_from_string(self.handler_key)
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(self.kdict, "tasks:*:handler"),
            )
            .unwrap()
        )
        return handler



from abc import ABCMeta, abstractmethod

from meiga import Result, NotImplementedMethodError

from petisco.legacy.application.pattern_base import PatternBase


class IRepository(PatternBase):

    __metaclass__ = ABCMeta

    @abstractmethod
    def save(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve_all(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def remove(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError



import traceback
from typing import List, Any

from meiga import Result, Failure
from meiga.decorators import meiga

from petisco.legacy.domain.aggregate_roots.info_id import InfoId
from petisco.legacy.domain.errors.critical_error import CriticalError
from petisco.legacy.domain.errors.unknown_error import UnknownError
from petisco.legacy.logger.interface_logger import ERROR, DEBUG
from petisco.legacy.logger.log_message import LogMessage
from petisco.legacy.notifier.domain.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.legacy.use_case.use_case import UseCase
from petisco.legacy.application.petisco import Petisco

DEFAULT_LOGGER = None
DEFAULT_NOTIFIER = None


class _UseCaseHandler:
    def __init__(
        self,
        logger=DEFAULT_LOGGER,
        logging_parameters_whitelist: List[str] = None,
        logging_types_blacklist: List[Any] = None,
        notifier=DEFAULT_NOTIFIER,
    ):
        self.logger = logger
        self.logging_parameters_whitelist = logging_parameters_whitelist
        self.logging_types_blacklist = logging_types_blacklist
        self.notifier = notifier
        self._check_logger()
        self._check_notifier()

    def _check_logger(self):
        if self.logger == DEFAULT_LOGGER:
            self.logger = Petisco.get_logger()

    def _check_notifier(self):
        if self.notifier == DEFAULT_NOTIFIER:
            self.notifier = Petisco.get_notifier()

    def __call__(self, cls):
        if not issubclass(cls, UseCase):
            raise TypeError("@use_case_handler only decorates UseCase subclasses")

        class UseCaseWrapped(cls):
            logger = self.logger
            logging_parameters_whitelist = self.logging_parameters_whitelist
            logging_types_blacklist = self.logging_types_blacklist
            notifier = self.notifier

            def execute(self, *args, **kwargs):
                info_id = kwargs.get("info_id")

                log_message = LogMessage(
                    layer="use_case", operation=f"{cls.__name__}", info_id=info_id
                )

                self.logger.log(DEBUG, log_message.set_message("Running Use Case"))

                if self.logging_parameters_whitelist:
                    loggable_kwargs = {}
                    for key, value in kwargs.items():
                        if key not in self.logging_parameters_whitelist:
                            continue
                        if getattr(value, "to_json", None):
                            loggable_kwargs[key] = value.to_json()
                        else:
                            loggable_kwargs[key] = value

                    if loggable_kwargs:
                        self.logger.log(DEBUG, log_message.set_message(loggable_kwargs))

                try:
                    result = self._run_execute(*args, **kwargs)
                except Exception as exception:
                    result = Failure(
                        UnknownError(
                            exception=exception,
                            input_parameters=kwargs if len(kwargs) > 0 else args,
                            executor=f"{cls.__name__} (Use Case)",
                            traceback=traceback.format_exc(),
                            filter_parameters=["headers"],
                        )
                    )

                if not isinstance(result, Result):
                    return result

                detail = getattr(result.value, "message", None)
                if detail:
                    detail = f"-> {detail}"
                else:
                    detail = ""

                if result.is_failure:
                    self.logger.log(
                        ERROR, log_message.set_message(f"{result} {detail}")
                    )
                    self.notify(result, info_id)
                else:
                    if not self._is_logging_type(result.value):
                        message = (
                            f"Success result of type: {type(result.value).__name__}"
                        )
                    else:
                        message = f"{result.value}"
                    self.logger.log(DEBUG, log_message.set_message(message))

                return result

            def _is_logging_type(self, value):
                if self.logging_types_blacklist:
                    for logging_type in self.logging_types_blacklist:
                        if isinstance(value, logging_type):
                            return False
                return True

            @meiga
            def _run_execute(self, *args, **kwargs) -> Result:
                return super().execute(*args, **kwargs)

            @meiga
            def notify(self, result, info_id: InfoId = None):
                if result.is_failure:
                    error = result.value
                    if issubclass(error.__class__, CriticalError):
                        self.notifier.publish(
                            NotifierExceptionMessage(
                                exception=error.exception,
                                executor=error.executor,
                                input_parameters=error.input_parameters,
                                traceback=error.traceback,
                                info_id=info_id,
                                info_petisco=Petisco.get_info(),
                            )
                        )

        return UseCaseWrapped


use_case_handler = _UseCaseHandler



import os

from petisco.legacy.persistence.pymongo.pymongo_persistence_config import (
    PyMongoPersistenceConfig,
)
from petisco.legacy.persistence.pymongo.pymongo_persistence_connector import (
    PyMongoPersistenceConnector,
)


def config_mongo_persistence():
    username = os.getenv("MONGODB_USERNAME")
    password = os.getenv("MONGODB_PASSWORD")
    port = int(os.getenv("MONGODB_PORT"))
    host = os.getenv("MONGODB_HOST")
    database = os.getenv("MONGODB_DATABASE")
    mongodb_config = PyMongoPersistenceConfig(
        host=host, user=username, password=password, port=port, database=database
    )
    mongodb_persistence_connector = PyMongoPersistenceConnector(config=mongodb_config)
    mongodb_persistence_connector.execute()



from petisco.legacy.domain.value_objects.value_object import ValueObject


class IntegerValueObject(ValueObject):
    def __init__(self, value: int):
        super(IntegerValueObject, self).__init__()
        self.value = value

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"[{self.__class__.__name__}: {self.value}]"

    def __eq__(self, other):
        if issubclass(other.__class__, self.__class__) or issubclass(
            self.__class__, other.__class__
        ):
            return self.value == other.value
        else:
            return False

    def to_str(self):
        return str(self.value)



from typing import Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class ConfigFramework:
    selected_framework: str
    config_file: str
    port: Optional[int] = 8080
    port_env: Optional[str] = None



import re

from pydantic import validator

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.model.value_object import ValueObject


class InvalidClientId(DomainEvent):
    def detail(self) -> str:
        return f"Invalid ClientId{self.detail()}"


class ClientId(ValueObject):
    _allow_utf8mb4: bool = True

    @validator("value")
    def validate_value(cls, value):
        if len(value) > 50:
            raise InvalidClientId(uuid_value=value)

        if not isinstance(value, str) or not re.search(
            r"^[\w]*(([',. -][\s]?[\w]?)?[\w]*)*$", value
        ):
            raise InvalidClientId(uuid_value=value)
        if not cls._allow_utf8mb4 and re.match("[^\u0000-\uffff]", value):
            raise InvalidClientId(uuid_value=value)

        return value.title()



from petisco.base.domain.message.command import Command
from petisco.base.domain.message.command_bus import CommandBus


class NotImplementedCommandBus(CommandBus):
    def dispatch(self, command: Command):
        self._check_is_command(command)
        meta = self.get_configured_meta()
        _ = command.update_meta(meta)

    def close(self):
        pass



from typing import List, Callable

from sqlalchemy.orm import scoped_session

from petisco.base.domain.persistence.interface_database import Database
from petisco.base.domain.persistence.persistence_models import PersistenceModels
from petisco.extra.sqlalchemy.sql.mysql.mysql_connection import MySqlConnection
from petisco.extra.sqlalchemy.sql.sql_session_scope_provider import (
    sql_session_scope_provider,
)


class MySqlDatabase(Database):
    @staticmethod
    def local_connection_checker():
        return MySqlDatabase(
            name="test", connection=MySqlConnection.create_local("test")
        )

    def __init__(
        self,
        name: str,
        connection: MySqlConnection,
        model_filename: str = None,
        print_sql_statements: bool = False,
        use_scoped_session: bool = True,
        use_future: bool = True,
    ):
        if not connection or not isinstance(connection, MySqlConnection):
            raise ConnectionError(
                "MySqlDatabase needs a valid MySqlConnection connection"
            )
        if model_filename:
            self.persistence_models = PersistenceModels.from_filename(model_filename)
        else:
            self.persistence_models = PersistenceModels(models={})
        self.connection = connection
        self.print_sql_statements = print_sql_statements
        self.use_scoped_session = use_scoped_session
        self.use_future = use_future

        super().__init__(name, self.persistence_models.models)
        self._init()

    def _init(self):
        from sqlalchemy.ext.declarative import declarative_base

        self.base = declarative_base()

    def create(self):
        self.persistence_models.import_models()
        from sqlalchemy import create_engine
        from sqlalchemy.orm import sessionmaker
        from sqlalchemy_utils import database_exists, create_database

        engine = create_engine(
            self.connection.url,
            pool_pre_ping=True,
            json_serializer=lambda obj: obj,
            json_deserializer=lambda obj: obj,
            echo=self.print_sql_statements,
            future=self.use_future,
        )

        if not database_exists(engine.url):
            create_database(engine.url)
            self.base.metadata.create_all(engine)

        self.session_maker = sessionmaker(bind=engine, future=self.use_future)

    def delete(self):
        pass
        # os.remove(self.connection.database_name)

    def clear_data(self):
        # Not available yet to avoid production problems.
        # This, needs to be well documented.
        # self.base.metadata.drop_all(self.engine)
        pass

    def is_available(self):
        try:
            session_scope = self.get_session_scope()
            with session_scope() as session:
                session.execute("SELECT 1")
                _is_available = True
        except Exception:  # noqa E722
            _is_available = False
        return _is_available

    def get_base(self):
        return self.base

    def get_model(self, model_name: str):
        model = self.persistence_models.get_imported_models().get(model_name)
        if not model:
            raise IndexError(
                f'Model "{model_name}" is not available for "{self.name}" database'
            )
        return model

    def get_model_names(self) -> List[str]:
        return list(self.persistence_models.get_models_names().keys())

    def get_session(self):
        if self.use_scoped_session:
            Session = scoped_session(self.session_maker)  # noqa
        else:
            Session = self.session_maker
        return Session

    def get_session_scope(self) -> Callable:
        return sql_session_scope_provider(self.get_session())



from typing import Any, Callable

from meiga import Result, Error, isSuccess, Success

from petisco.legacy import Persistence

from tests.end2end.flask_app.users_count.domain.interface_users_count_repository import (
    IUsersCountRepository,
)


class SqlUsersCountRepository(IUsersCountRepository):
    @staticmethod
    def build():
        return SqlUsersCountRepository(
            session_scope=Persistence.get_session_scope("petisco-sql"),
            users_count_model=Persistence.get_model("petisco-sql", "users_count"),
        )

    def __init__(self, session_scope: Callable, users_count_model: Any):
        self.session_scope = session_scope
        self.UsersCountModel = users_count_model

    def increase(self) -> Result[bool, Error]:
        with self.session_scope() as session:
            users_count_model = session.query(self.UsersCountModel).first()
            if not users_count_model:
                users_count_model = self.UsersCountModel(count=1)
                session.add(users_count_model)
                return isSuccess

            users_count_model.count += 1

            return isSuccess

    def decrease(self) -> Result[bool, Error]:
        with self.session_scope() as session:
            users_count_model = session.query(self.UsersCountModel).first()
            if not users_count_model:
                users_count_model = self.UsersCountModel(count=0)
                session.add(users_count_model)
                return isSuccess

            users_count_model.count -= 1

            return isSuccess

    def count(self) -> Result[int, Error]:
        with self.session_scope() as session:
            users_count_model = session.query(self.UsersCountModel).first()
            if not users_count_model:
                return Success(0)

            return Success(users_count_model.count)



from typing import Callable, List, Optional

from pydantic import BaseSettings, Field

from petisco import __version__
from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.application.dependency_injection.injector import Injector
from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.domain.message.chaos.service_deployed import ServiceDeployed
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.extra.rabbitmq.dependencies import get_default_message_dependencies
from petisco.extra.slack.dependencies import get_default_notifier_dependencies


class Application(BaseSettings):
    name: str
    version: str
    organization: str
    environment: str = Field("local", env="ENVIRONMENT")
    dependencies_provider: Optional[Callable[..., List[Dependency]]] = lambda: []
    configurers: Optional[List[ApplicationConfigurer]] = []

    def configure(self, testing: bool = False):

        before_dependecies_configurers = [
            configurer
            for configurer in self.configurers
            if configurer.execute_after_dependencies is False
        ]
        for configurer in before_dependecies_configurers:
            configurer.execute(testing)

        Injector.set_dependencies(self.get_dependencies())

        after_dependencies_configurers = [
            configurer
            for configurer in self.configurers
            if configurer.execute_after_dependencies is True
        ]
        for configurer in after_dependencies_configurers:
            configurer.execute(testing)

    def get_dependencies(self) -> List[Dependency]:
        default_dependencies = (
            get_default_message_dependencies() + get_default_notifier_dependencies()
        )
        default_dependencies_dict = {
            dependency.name: dependency for dependency in default_dependencies
        }
        provided_dependencies = self.dependencies_provider()
        given_dependencies_dict = {
            dependency.name: dependency for dependency in provided_dependencies
        }
        merged_dependecies = {**default_dependencies_dict, **given_dependencies_dict}
        return list(merged_dependecies.values())

    def clear(self):
        Injector.clear()

    def info(self):
        info = self.dict()
        info["dependencies"] = {
            dependency.name: dependency.get_instance().info()
            for dependency in self.get_dependencies()
        }
        del info["dependencies_provider"]
        del info["configurers"]
        return info

    def publish_deploy_event(self, domain_event: Optional[DomainEvent] = None):
        try:
            domain_event_bus: DomainEventBus = Injector.get("domain_event_bus")
            domain_event_bus.publish(
                ServiceDeployed(app_name=self.name, app_version=self.version)
            )
        except:  # noqa
            raise TypeError(
                'To publish an event to the domain event bus, please add a dependency with name "domain_event_bus" on Application dependencies'
            )

    def notify_deploy(self):
        try:
            notifier: Notifier = Injector.get("notifier")
            notifier.publish(
                NotifierMessage(
                    title=f":rocket: {self.name} is deployed",
                    meta={
                        "application": f"{self.name} (v{self.version.rstrip()})",
                        "petisco": __version__,
                        "environment": self.environment,
                    },
                )
            )
        except:  # noqa
            raise TypeError(
                'To notify the deploy, please add a dependency with name "notifier" on Application dependencies'
            )



import inspect
from typing import Dict, TypeVar, Callable

from deprecation import deprecated

from petisco.legacy.application.interface_app_service import IAppService
from petisco.legacy.application.singleton import Singleton

AppService = TypeVar("AppService", bound=IAppService)


class AppServices(metaclass=Singleton):
    def __init__(self, app_services: Dict[str, AppService] = None):
        self._app_services = app_services if app_services else {}

    def get_info(self):
        return {
            name: app_service.info() for name, app_service in self._app_services.items()
        }

    @staticmethod
    def info():
        return AppServices.get_instance().get_info()

    @staticmethod
    def load(provider: Callable):
        repositories = provider()
        return AppServices(repositories)

    @staticmethod
    def get_instance():
        try:
            return AppServices()
        except Exception as e:  # noqa E722
            frame_info = inspect.stack()[1]
            raise ImportError(
                f"AppServices must be configured. If not, you cannot obtain app serives\n"
                f"Following code must be executed after AppServices initialization:\n"
                f"\tfilename: {frame_info.filename}\n"
                f"\tlineno: {frame_info.lineno}\n"
                f"\tfunction: {frame_info.function}\n"
                f"\tcode_context: {frame_info.code_context}\n\n"
            )

    def add(self, name: str, app_service: AppService):
        if name in self._app_services:
            raise NameError(f"AppService {name} is already added to AppServices")
        self._app_services[name] = app_service

    def remove(self, name: str):
        if name in self._app_services:
            del self._app_services[name]
        else:
            raise IndexError(f"AppService cannot be removed. {name} not exists")

    @staticmethod
    def get(name: str) -> AppService:
        app_services = AppServices.get_instance()._app_services
        if app_services is None:
            raise ValueError(
                "AppServices: no app service has been declared. Please, initialize it (i.e AppServices.from_provider(provider_func))"
            )
        app_service = app_services.get(name)
        if not app_service:
            raise ValueError(
                f"AppServices: {name} app service is not defined.  Please, add it (i.e AppServices.from_provider(provider_func))"
            )
        return app_service

    @staticmethod
    @deprecated("This method is deprecated. Please, use AppServices.load")
    def from_provider(provider: Callable):
        return AppServices.load(provider)



import os
import random
from time import sleep
from typing import Optional, Dict, List

from pika.adapters.blocking_connection import BlockingChannel
from pika.spec import Basic

from petisco.base.domain.message.chaos.message_chaos import MessageChaos

MESSAGE_CHAOS_PERCENTAGE_SIMULATE_NACK_KEY = "MESSAGE_CHAOS_PERCENTAGE_SIMULATE_NACK"
MESSAGE_CHAOS_DELAY_BEFORE_EVENT_HANDLER_SECONDS_KEY = (
    "MESSAGE_CHAOS_DELAY_BEFORE_EVENT_HANDLER_SECONDS"
)
MESSAGE_CHAOS_PERCENTAGE_SIMULATE_FAILURES_KEY = (
    "MESSAGE_CHAOS_PERCENTAGE_SIMULATE_FAILURES"
)

MESSAGE_CHAOS_PROTECTED_ROUTING_KEYS_KEY = "MESSAGE_CHAOS_PROTECTED_ROUTING_KEYS"


class RabbitMqMessageChaos(MessageChaos):
    def __init__(
        self,
        percentage_simulate_nack: Optional[float] = None,
        delay_before_event_handler_second: Optional[float] = None,
        percentage_simulate_failures: Optional[float] = None,
        protected_routing_keys: Optional[List[str]] = None,
    ):
        """
        Parameters
        ----------
        percentage_simulate_nack
            Percentage of simulate nack [0.0 -> 1.0]. Where 1.0 rejects all the event.
        delay_before_event_handler_second
            Delay event handler execution for a given number of seconds.
        percentage_simulate_failures
            Percentage of simulate failures [0.0 -> 1.0]. Where 1.0 simulate always a failure on handlers.
        protected_routing_keys
            Routing keys where chaos will not be applied
        """
        self._set_percentage_simulate_nack(percentage_simulate_nack)
        self._set_delay_before_even_handler_second(delay_before_event_handler_second)
        self._set_percentage_simulate_failures(percentage_simulate_failures)
        self._set_protected_routing_keys(protected_routing_keys)

    def _set_percentage_simulate_nack(self, percentage_simulate_nack):
        if percentage_simulate_nack is None:
            percentage_simulate_nack = os.environ.get(
                MESSAGE_CHAOS_PERCENTAGE_SIMULATE_NACK_KEY
            )
        self.percentage_simulate_nack = self._float(percentage_simulate_nack)

    def _set_delay_before_even_handler_second(self, delay_before_even_handler_second):
        if delay_before_even_handler_second is None:
            delay_before_even_handler_second = os.environ.get(
                MESSAGE_CHAOS_DELAY_BEFORE_EVENT_HANDLER_SECONDS_KEY
            )
        self.delay_before_even_handler_second = self._float(
            delay_before_even_handler_second
        )

    def _set_percentage_simulate_failures(self, percentage_simulate_failures):
        if percentage_simulate_failures is None:
            percentage_simulate_failures = os.environ.get(
                MESSAGE_CHAOS_PERCENTAGE_SIMULATE_FAILURES_KEY
            )
        self.percentage_simulate_failures = self._float(percentage_simulate_failures)

    def _set_protected_routing_keys(self, protected_routing_keys):
        if protected_routing_keys is None:
            protected_routing_keys = os.environ.get(
                MESSAGE_CHAOS_PROTECTED_ROUTING_KEYS_KEY
            )
        self.protected_routing_keys = self._list(protected_routing_keys)

    def _float(self, value):
        return float(value) if value is not None else None

    def _list(self, value):
        if value is None:
            return None

        if isinstance(value, str):
            value = list(value.split(","))
        return value

    def __repr__(self):
        return f"RabbitMqEventChaos: {self.info()}"

    def info(self) -> Dict:
        return {
            MESSAGE_CHAOS_PERCENTAGE_SIMULATE_NACK_KEY: self.percentage_simulate_nack,
            MESSAGE_CHAOS_DELAY_BEFORE_EVENT_HANDLER_SECONDS_KEY: self.delay_before_even_handler_second,
            MESSAGE_CHAOS_PERCENTAGE_SIMULATE_FAILURES_KEY: self.percentage_simulate_failures,
            MESSAGE_CHAOS_PROTECTED_ROUTING_KEYS_KEY: self.protected_routing_keys,
        }

    def nack_simulation(self, ch: BlockingChannel, method: Basic.Deliver) -> bool:
        routing_key = method.routing_key if method else None

        if (self.percentage_simulate_nack is None) or (
            random.random() > self.percentage_simulate_nack
        ):
            return False

        if (
            self.protected_routing_keys is None
            or routing_key not in self.protected_routing_keys
        ):
            ch.basic_nack(delivery_tag=method.delivery_tag)
            return True
        else:
            return False

    def failure_simulation(self, method: Basic.Deliver) -> bool:
        routing_key = method.routing_key if method else None

        if (self.percentage_simulate_failures is None) or (
            random.random() > self.percentage_simulate_failures
        ):
            return False
        else:
            return (
                True
                if self.protected_routing_keys is None
                or routing_key not in self.protected_routing_keys
                else False
            )

    def delay(self):
        if self.delay_before_even_handler_second is None:
            pass
        else:
            sleep(self.delay_before_even_handler_second)



from typing import Optional

from dataclasses import dataclass

from petisco.legacy.application.config.events.config_event_publisher import (
    ConfigEventsPublisher,
)
from petisco.legacy.application.config.events.config_event_subscriber import (
    ConfigEventsSubscriber,
)


@dataclass
class ConfigEvents:
    publish_deploy_event: Optional[bool] = False
    config_event_publisher: Optional[ConfigEventsPublisher] = ConfigEventsPublisher()
    config_event_subscriber: Optional[ConfigEventsSubscriber] = ConfigEventsSubscriber()

    @staticmethod
    def from_dict(kdict):
        if not kdict or not isinstance(kdict, dict):
            return ConfigEvents()

        config_event_publisher = ConfigEventsPublisher()
        config_event_publisher_dict = kdict.get("publisher")
        if config_event_publisher_dict:
            config_event_publisher = ConfigEventsPublisher.from_dict(
                config_event_publisher_dict
            )

        config_event_subscriber = ConfigEventsSubscriber()
        config_event_subscriber_dict = kdict.get("subscriber")
        if config_event_subscriber_dict:
            config_event_subscriber = ConfigEventsSubscriber.from_dict(
                config_event_subscriber_dict
            )

        return ConfigEvents(
            publish_deploy_event=kdict.get("publish_deploy_event"),
            config_event_publisher=config_event_publisher,
            config_event_subscriber=config_event_subscriber,
        )



from typing import Tuple

from abc import ABCMeta, abstractmethod


DEFAULT_HTTP_ERROR_MESSAGE = "Unknown Error"


class HttpError:

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(
        self,
        message: str = DEFAULT_HTTP_ERROR_MESSAGE,
        code: int = 500,
        type_error: str = None,
    ):
        self.message = message
        self.code = code
        self.type_error = type_error if type_error else self.__class__.__name__

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.code} | {self.message}"

    def handle(self) -> Tuple[dict, int]:
        return (
            {"error": {"type": self.type_error, "message": self.message}},
            self.code,
        )



from typing import Any, Callable

from meiga import Result, Error, isSuccess, Success, Failure

from petisco.legacy import Persistence
from petisco.legacy.domain.value_objects.client_id import ClientId
from petisco.legacy.domain.value_objects.name import Name
from petisco.legacy.domain.value_objects.user_id import UserId
from tests.end2end.flask_app.shared.domain.repositories.interface_user_repository import (
    IUserRepository,
)
from tests.end2end.flask_app.shared.domain.repositories.user_already_exist_error import (
    UserAlreadyExistError,
)
from tests.end2end.flask_app.shared.domain.repositories.user_not_found_error import (
    UserNotFoundError,
)
from tests.end2end.flask_app.user.domain.aggregate_roots.user import User


class SqlUserRepository(IUserRepository):
    @staticmethod
    def build():
        return SqlUserRepository(
            session_scope=Persistence.get_session_scope("petisco-sql"),
            user_model=Persistence.get_model("petisco-sql", "user"),
        )

    def __init__(self, session_scope: Callable, user_model: Any):
        self.session_scope = session_scope
        self.UserModel = user_model

    def save(self, user: User) -> Result[bool, Error]:

        result = self.exists(user.user_id)
        if result.is_success:
            return Failure(UserAlreadyExistError(user.user_id))

        with self.session_scope() as session:

            user = self.UserModel(
                client_id=user.client_id.value,
                user_id=user.user_id.value,
                name=user.name.value,
            )
            session.add(user)
        return isSuccess

    def retrieve(self, client_id: ClientId, user_id: UserId) -> Result[User, Error]:

        with self.session_scope() as session:
            user_model = (
                session.query(self.UserModel)
                .filter(self.UserModel.client_id == client_id.value)
                .filter(self.UserModel.user_id == user_id.value)
                .first()
            )
            if not user_model:
                return Failure(UserNotFoundError(user_id))

            return Success(
                User(
                    name=Name(user_model.name),
                    client_id=ClientId(user_model.client_id),
                    user_id=UserId(user_model.user_id),
                )
            )

    def exists(self, user_id: UserId) -> Result[bool, Error]:
        with self.session_scope() as session:
            user = (
                session.query(self.UserModel)
                .filter(self.UserModel.user_id == user_id.value)
                .first()
            )
            if user:
                return isSuccess
            return Failure(UserNotFoundError(user_id))



from typing import Any

from pydantic import validator
from pydantic.main import BaseModel

from petisco.base.domain.errors.defaults.invalid_value_object import InvalidValueObject


class ValueObject(BaseModel):
    value: Any

    def __init__(self, value: Any, **data) -> None:
        super().__init__(value=value, **data)

    def dict(self, **kwargs):
        return self.value

    def __setattr__(self, name, value):
        raise TypeError("ValueObject objects are immutable")

    @classmethod
    def from_value(cls, value):
        return cls(value=value)

    @validator("value")
    def validate_value(cls, value):
        if value is None:
            raise InvalidValueObject()
        return value



from dataclasses import dataclass

from petisco.legacy.event.shared.domain.event import Event
from petisco.legacy.webhooks.webhook.domain.webhook_id import WebhookId


@dataclass
class WebhookCreated(Event):
    webhook_id: WebhookId
    active: bool
    subscribed_event_name: str
    subscribed_event_version: str

    def __init__(
        self,
        webhook_id: WebhookId,
        active: bool,
        subscribed_event_name: str,
        subscribed_event_version: str,
    ):
        self.webhook_id = webhook_id
        self.active = active
        self.subscribed_event_name = subscribed_event_name
        self.subscribed_event_version = subscribed_event_version
        super().__init__()



import logging
from typing import Dict

from fastapi import HTTPException
from meiga import Result

from petisco.base.application.controller.http_error import (
    DEFAULT_HTTP_ERROR_DETAIL,
    HttpError,
)
from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.errors.unknown_error import UnknownError

logger = logging.getLogger()


def fastapi_failure_handler(result: Result, error_map: Dict[type, HttpError]):
    domain_error = result.value
    error_type = type(domain_error)
    http_error = error_map.get(error_type, HttpError())

    if isinstance(result.value, UnknownError):
        logger.error(result.value.__dict__)
    elif error_type not in error_map:
        error_info = {
            "message": f"Error '{result.value.__class__.__name__}' is not mapped in controller"
        }
        logger.error(error_info)

    detail = "Unknown Error"
    if isinstance(domain_error, DomainError):
        detail = (
            http_error.detail
            if http_error.detail != DEFAULT_HTTP_ERROR_DETAIL
            else domain_error.detail()
        )
    raise HTTPException(status_code=http_error.status_code, detail=detail)



import json
from abc import abstractmethod
from typing import Optional, List, Dict

from dataclasses import dataclass

from petisco.legacy.domain.aggregate_roots.info_id import InfoId


@dataclass
class NotifierMessage:
    title: str = None
    message: str = None
    info_petisco: Optional[Dict] = None
    info_id: Optional[InfoId] = None
    files: Optional[List[Dict]] = None

    @abstractmethod
    def __str__(self) -> str:
        return f"Title: {self.title}\nMessage: {self.message}\nPetisco info: {json.dumps(self.info_petisco)}\nInfo Id: {json.dumps(self.info_id)}"



from abc import abstractmethod

from meiga import NotImplementedMethodError, Result

from petisco.base.application.patterns.base_pattern import BasePattern


class Repository(BasePattern):
    @abstractmethod
    def save(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve_all(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError

    @abstractmethod
    def remove(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError



import threading
import traceback
from time import sleep
from typing import Dict

from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import StreamLostError

from petisco.legacy.event.legacy.rabbitmq.create_exchange_and_bind_queue import (
    create_exchange_and_bind_queue,
    create_dead_letter_exchange_and_bind_queue,
)
from petisco.legacy.event.legacy.rabbitmq.get_event_binding_key import (
    get_event_binding_key,
)
from petisco.legacy.event.shared.infrastructure.rabbitmq.rabbitmq_connector import (
    RabbitMqConnector,
)
from petisco.legacy.event.legacy.subscriber.domain.config_event_subscriber import (
    ConfigEventSubscriber,
)
from petisco.legacy.event.legacy.subscriber.domain.interface_event_subscriber import (
    IEventSubscriber,
)


class RabbitMQEventSubscriber(IEventSubscriber):
    def __init__(
        self,
        connector: RabbitMqConnector,
        subscribers: Dict[str, ConfigEventSubscriber] = None,
        connection_name: str = "subscriber",
    ):
        if not connector:
            raise TypeError("RabbitMQEventSubscriber: Invalid Given RabbitMQConnector")

        self._connector = connector
        self._connection = None
        self._connection_name = connection_name
        self._channels = {}
        self._thread = None
        super().__init__(subscribers)

    def start(self):
        if not self.subscribers:
            raise RuntimeError(
                "RabbitMQEventSubscriber: cannot start consuming event without any subscriber defined"
            )
        self._thread = threading.Thread(target=self._start)
        self._thread.start()

    def _start(self):
        self._connect()
        self._create_channels()
        self._bind_queue_and_subscribers()
        self._start_consuming()

    def _connect(self):
        self._connection = self._connector.get_connection(self._connection_name)

    def _check_connection(self):
        if not self._connection.is_open:
            self._connect()

    def _get_channel(self) -> BlockingChannel:
        self._check_connection()
        try:
            channel = self._connection.channel()
        except StreamLostError:
            self._check_connection()
            channel = self._connection.channel()
        return channel

    def _create_channels(self):
        # This uses the basic.qos protocol method to tell RabbitMQ not to give more than one message to a worker at a
        # time. Or, in other words, don't dispatch a new message to a worker until it has processed and acknowledged
        # the previous one. Instead, it will dispatch it to the next worker that is not still busy.
        for name in self.subscribers.keys():
            self._channels[name] = self._get_channel()
            self._channels[name].basic_qos(prefetch_count=1)

    def _bind_queue_and_subscribers(self):
        for name, subscriber_config in self.subscribers.items():
            self._setup_exchanges_and_queues(subscriber_config, self._channels[name])
            queue = (
                subscriber_config.topic
                if not subscriber_config.dead_letter
                else f"dl-{subscriber_config.topic}"
            )
            self._channels[name].basic_consume(
                queue=queue, on_message_callback=subscriber_config.get_handler()
            )

    def _setup_exchanges_and_queues(
        self, subscriber_config: ConfigEventSubscriber, channel: BlockingChannel
    ):
        exchange = subscriber_config.service
        queue = subscriber_config.topic
        binding_key = get_event_binding_key(
            subscriber_config.organization, subscriber_config.service
        )

        create_dead_letter_exchange_and_bind_queue(
            channel=channel, exchange=exchange, queue=queue, binding_key=binding_key
        )
        create_exchange_and_bind_queue(
            channel=channel,
            exchange=exchange,
            queue=queue,
            binding_key=binding_key,
            dead_letter=True,
        )

    def _start_consuming(self):
        for name in self.subscribers.keys():
            self._channels[name].start_consuming()

    def get_subscribers_status(self) -> Dict[str, str]:
        subscribers_status = {}
        for name, channel in self._channels.items():
            subscribers_status[name] = (
                "subscribed" if len(channel.consumer_tags) > 0 else "unsubscribed"
            )
        return subscribers_status

    def _unsubscribe_all(self):
        def _stop_consuming_channels():
            for name in self.subscribers.keys():
                if name in self._channels:
                    self._channels[name].stop_consuming()
                    self._channels[name].cancel()

        def _await_for_stop_consuming_channels():
            sleep(2.0)

        self._connection.call_later(0, _stop_consuming_channels)
        _await_for_stop_consuming_channels()

    def stop(self):
        def _log_stop_exception(e: Exception):
            from petisco.legacy import LogMessage, ERROR, Petisco

            logger = Petisco.get_logger()
            log_message = LogMessage(
                layer="petisco", operation="RabbitMQEventSubscriber"
            )
            message = f"Error stopping RabbitMQEventSubscriber: {repr(e.__class__)} {e} | {traceback.format_exc()}"
            logger.log(ERROR, log_message.set_message(message))

        if self._thread and self._thread.is_alive():
            self._unsubscribe_all()
            try:
                self._thread.join()
                self._thread = None
            except Exception as e:
                _log_stop_exception(e)

    def info(self) -> Dict:
        is_open = False
        if self._connection:
            is_open = self._connection.is_open
        return {
            "name": self.__class__.__name__,
            "connection.is_open": is_open,
            "subscribers_status": self.get_subscribers_status(),
        }



from typing import Dict

from meiga import Result, Error, Failure, Success

from petisco.legacy.application.petisco import Petisco
from petisco.legacy.persistence.persistence import Persistence
from petisco.legacy.use_case.use_case import UseCase
from petisco.legacy.modules.healthcheck.domain.persistence_error import PersistenceError


class HealthcheckProvider(UseCase):
    def execute(self, petisco: Petisco) -> Result[Dict, Error]:

        healthcheck = {
            "app_name": petisco._app_name,
            "app_version": petisco._app_version,
            "persistence_available": Persistence.is_available(),
        }

        if Persistence.exist() and not Persistence.is_available():
            return Failure(PersistenceError("Persistence is not available"))

        return Success(healthcheck)



from petisco.legacy.logger.log_message import LogMessage
from petisco.legacy.logger.interface_logger import ILogger


class NotImplementedLogger(ILogger):
    def log(self, logging_level, log_message: LogMessage):
        pass



from petisco.legacy.domain.value_objects.string_value_object import StringValueObject
from petisco.legacy.domain.errors.given_name_is_not_valid_error import (
    GivenNameIsNotValidError,
)


class Name(StringValueObject):
    def __init__(self, value: str):
        self.value = value
        super(Name, self).__init__(self.value)

    def guard(self):
        self._ensure_value_contains_valid_char(
            raise_cls=GivenNameIsNotValidError, allow_utf8mb4=False
        )
        self._ensure_value_is_less_than_n_char(50)
        self._ensure_value_is_greater_than_n_char(0)



from fastapi import HTTPException


def assert_http_exception(
    http_exception_current: HTTPException, http_exception_expected: HTTPException
):
    assert (
        http_exception_current.status_code == http_exception_expected.status_code
        and http_exception_current.detail == http_exception_expected.detail
    ), f"| status_code: {http_exception_current.status_code} - {http_exception_expected.status_code} | -> | detail: {http_exception_current.detail} - {http_exception_expected.detail}|"



import json
from datetime import datetime
from meiga import Result

from petisco.legacy.domain.date_parser import DateParser
from petisco.legacy.webhooks.webhook.domain.webhook_id import WebhookId
from petisco.legacy.webhooks.webhook.domain.webhook import Webhook
from petisco.legacy.domain.aggregate_roots.aggregate_root import AggregateRoot
from petisco.legacy.webhooks.webhook.domain.webhook_sender_failed import (
    WebhookSenderFailed,
)
from petisco.legacy.webhooks.webhook.domain.webhook_result_id import WebhookResultId
from petisco.legacy.webhooks.webhook.domain.webhook_request_result import (
    WebhookRequestResult,
)
from petisco.legacy.webhooks.webhook.domain.webhook_response_result import (
    WebhookResponseResult,
)


class WebhookResult(AggregateRoot):
    @staticmethod
    def create(
        webhook: Webhook,
        webhook_result_id: WebhookResultId,
        sent_on: datetime,
        request_headers: dict,
        request_body: dict,
        result: Result,
    ):
        response = WebhookResponseResult.from_result(result)
        request = WebhookRequestResult(request_headers, request_body)
        webhook_result = WebhookResult(
            webhook_result_id,
            webhook.webhook_id,
            sent_on,
            response,
            request,
            result.is_success,
        )

        if not webhook_result.is_success:
            webhook_result.record(
                WebhookSenderFailed(
                    webhook_id=webhook.webhook_id, webhook_result_id=webhook_result_id
                )
            )

        return webhook_result

    @staticmethod
    def from_dict(kdict: dict):
        return WebhookResult(
            webhook_result_id=WebhookResultId(kdict.get("webhook_result_id")),
            webhook_id=WebhookId(kdict.get("webhook_id")),
            sent_on=DateParser.datetime_from_str(kdict.get("sent_on")),
            response=WebhookResponseResult.from_dict(kdict.get("response"))
            if kdict.get("response")
            else None,
            request=WebhookRequestResult.from_dict(kdict.get("request"))
            if kdict.get("request")
            else None,
            is_success=kdict.get("is_success"),
        )

    def to_dict(self):
        return {
            "webhook_result_id": self.webhook_result_id.value,
            "webhook_id": self.webhook_id.value,
            "sent_on": DateParser.str_from_datetime(self.sent_on),
            "response": self.response.to_dict() if self.response else None,
            "request": self.request.to_dict() if self.request else None,
            "is_success": self.is_success,
        }

    def __init__(
        self,
        webhook_result_id: WebhookResultId,
        webhook_id: WebhookId,
        sent_on: datetime,
        response: WebhookResponseResult,
        request: WebhookRequestResult,
        is_success: bool,
    ):
        self.webhook_result_id = webhook_result_id
        self.webhook_id = webhook_id
        self.sent_on = sent_on
        self.response = response
        self.request = request
        self.is_success = is_success
        super().__init__()

    def __repr__(self):
        return json.dumps(self.to_dict())

    def __eq__(self, other):
        if issubclass(other.__class__, self.__class__) or issubclass(
            self.__class__, other.__class__
        ):
            return self.to_dict() == other.to_dict()
        else:
            return False



from dataclasses import dataclass

from petisco.legacy.webhooks.webhook.domain.webhook_result_id import WebhookResultId
from petisco.legacy.event.shared.domain.event import Event
from petisco.legacy.webhooks.webhook.domain.webhook_id import WebhookId


@dataclass
class WebhookSenderFailed(Event):
    webhook_id: WebhookId
    webhook_result_id: WebhookResultId

    def __init__(self, webhook_id: WebhookId, webhook_result_id: WebhookResultId):
        self.webhook_id = webhook_id
        self.webhook_result_id = webhook_result_id
        super().__init__()



from petisco.legacy.domain.errors.critical_error import CriticalError


class PyMongoOperationalDatabaseError(CriticalError):
    pass


class PyMongoInvalidDatabaseNameError(CriticalError):
    pass


class PyMongoInvalidCollectionNameError(CriticalError):
    pass



import re

from petisco.legacy.event.shared.domain.event_subscriber import EventSubscriber


class RabbitMqEventSubscriberQueueNameFormatter:
    @staticmethod
    def format(subscriber: EventSubscriber, exchange_name: str = None) -> str:
        event_name = (
            re.sub(r"(?<!^)(?=[A-Z])", "_", subscriber.event_name)
            .lower()
            .replace(".", "_")
        )
        event_name = f"{subscriber.event_version}.event.{event_name}"
        return f"{exchange_name}.{event_name}" if exchange_name else event_name

    @staticmethod
    def format_retry(subscriber: EventSubscriber, exchange_name: str = None) -> str:
        queue_name = RabbitMqEventSubscriberQueueNameFormatter.format(
            subscriber, exchange_name
        )
        return f"retry.{queue_name}"

    @staticmethod
    def format_dead_letter(
        subscriber: EventSubscriber, exchange_name: str = None
    ) -> str:
        queue_name = RabbitMqEventSubscriberQueueNameFormatter.format(
            subscriber, exchange_name
        )
        return f"dead_letter.{queue_name}"



from typing import List

from petisco.legacy.event.shared.domain.event import Event
from petisco.legacy.event.shared.domain.event_subscriber import EventSubscriber
from petisco.legacy.event.configurer.domain.interface_event_configurer import (
    IEventConfigurer,
)


class NotImplementedEventConfigurer(IEventConfigurer):
    def configure(self):
        self.configure_subscribers([])

    def configure_event(self, event: Event):
        self.configure_subscribers(
            [
                EventSubscriber(
                    event_name=event.event_name,
                    event_version=event.event_version,
                    handlers=[],
                )
            ]
        )

    def configure_subscribers(self, subscribers: List[EventSubscriber]):
        pass

    def clear(self):
        pass



from petisco.extra.elastic.is_elastic_available import is_elastic_available

rabbitmq = []
if is_elastic_available():

    from petisco.extra.elastic.elastic_connection import ElasticConnection
    from petisco.extra.elastic.elastic_database import ElasticDatabase

    elastic = ["ElasticConnection", "ElasticDatabase"]
else:
    elastic = []

__all__ = elastic



from petisco.legacy import Petisco, Persistence, SqliteDatabase, SqliteConnection
import os
import pytest

# from tests.end2end.flask_app import petisco_config
# petisco_config()
from tests.end2end.flask_app.config.app_services import load_app_services
from tests.end2end.flask_app.config.repositories import load_repositories

username = os.getenv("MONGODB_USERNAME")
password = os.getenv("MONGODB_PASSWORD")
port = (
    int(os.getenv("MONGODB_PORT"))
    if os.getenv("MONGODB_PORT")
    else os.getenv("MONGODB_PORT")
)
host = os.getenv("MONGODB_HOST")
database = os.getenv("MONGODB_DATABASE")


@pytest.fixture
def given_any_apikey():
    apikey = "apikey"
    return apikey


@pytest.fixture
def given_code_injection_name():
    return "<script>evil()</script>"


@pytest.fixture
def petisco_yml_path_flask_app():
    return f"{os.path.dirname(os.path.abspath(__file__))}/flask_app"


@pytest.fixture
def petisco_yml_path_end2end():
    return f"{os.path.dirname(os.path.abspath(__file__))}/ymls"


@pytest.fixture
def given_petisco_flask_app(petisco_yml_path_flask_app):
    Petisco.clear()
    petisco = Petisco.from_filename(f"{petisco_yml_path_flask_app}/petisco.yml")
    petisco.configure_events(petisco_yml_path_flask_app + "/petisco.events.yml")

    load_app_services()
    sql_database = SqliteDatabase(
        name="petisco-sql",
        connection=SqliteConnection.create("sqlite", "petisco-test.db"),
        model_filename=f"{petisco_yml_path_flask_app}/petisco.sql.models.yml",
    )
    persistence = Persistence()
    persistence.add(sql_database, skip_if_exist=True)
    persistence.create()
    load_repositories()

    yield petisco

    persistence.clear_data()

    Petisco.clear()
    petisco.stop()


@pytest.fixture
def client_app(given_petisco_flask_app):
    app = given_petisco_flask_app.get_app()
    with app.app.test_client() as c:
        yield c


@pytest.fixture
def given_petisco_flask_app_with_mongodb(petisco_yml_path_flask_app):
    Petisco.clear()
    petisco = Petisco.from_filename(
        f"{petisco_yml_path_flask_app}/petisco_with_mongo.yml"
    )
    petisco.configure_events(petisco_yml_path_flask_app + "/petisco.events.yml")
    yield petisco
    Petisco.clear()
    petisco.stop()


@pytest.fixture
def client_app_with_mongodb(given_petisco_flask_app_with_mongodb):
    app = given_petisco_flask_app_with_mongodb.get_app()
    with app.app.test_client() as c:
        yield c



from typing import Dict

from petisco.legacy import IRepository, Repositories
from tests.end2end.flask_app.shared.infrastructure.repositories.mongodb_user_repository import (
    MongoDBUserRepository,
)
from tests.end2end.flask_app.shared.infrastructure.repositories.sql_user_repository import (
    SqlUserRepository,
)
from tests.end2end.flask_app.users_count.infrastructure.inmemory_users_count_repository import (
    InMemoryUsersCountRepository,
)
from tests.end2end.flask_app.users_count.infrastructure.sql_users_count_repository import (
    SqlUsersCountRepository,
)


def load_repositories(with_mongo: bool = False):
    if with_mongo:
        Repositories.load(repositories_provider_mongo)
    else:
        Repositories.load(repositories_provider)


def repositories_provider() -> Dict[str, IRepository]:
    return {
        "user": SqlUserRepository.build(),
        "users_count": SqlUsersCountRepository.build(),
    }


def repositories_provider_mongo() -> Dict[str, IRepository]:
    return {
        "user": MongoDBUserRepository.build(),
        "users_count": InMemoryUsersCountRepository(),
    }



import functools

import pytest

from petisco.legacy.event.shared.infrastructure.rabbitmq.rabbitmq_is_running_locally import (
    rabbitmq_is_running_locally,
)
from petisco.legacy.persistence.elastic.elastic_is_running_locally import (
    elastic_is_running_locally,
)
from petisco.legacy.persistence.sql.mysql_is_running_locally import (
    mysql_is_running_locally,
)


def testing_with_rabbitmq(func):
    @pytest.mark.skipif(
        not rabbitmq_is_running_locally(), reason="RabbitMQ is not running locally"
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def testing_with_mysql(func):
    @pytest.mark.skipif(
        not mysql_is_running_locally(), reason="MySQL is not running locally"
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def testing_with_elastic(func):
    @pytest.mark.skipif(
        not elastic_is_running_locally(), reason="Elastic is not running locally"
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper



import base64
import json

from meiga import Result, Error, Failure, Success
from meiga.decorators import meiga

from petisco.legacy.security.token_decoder.interface_token_decoder import ITokenDecoder
from petisco.legacy.security.token_decoder.invalid_token_error import InvalidTokenError
from petisco.legacy.security.token_decoder.token import Token


class TokenDecoder(ITokenDecoder):
    @meiga
    def execute(self, auth_token) -> Result[Token, Error]:
        token_payload = self.decode_token_payload(auth_token).unwrap_or_return()
        return Success(Token.from_token_payload(token_payload))

    def decode_token_payload(self, token):
        try:
            token_payload = token.split(".")[1]
            token_payload += "=" * ((4 - len(token_payload) % 4) % 4)
            token_payload = json.loads(base64.b64decode(token_payload).decode("utf-8"))
        except (IndexError, AttributeError):
            return Failure(InvalidTokenError())

        return Success(token_payload)



from typing import List

from petisco.base.domain.persistence.interface_database import Database
from petisco.base.domain.persistence.persistence_models import PersistenceModels


class FakeDatabase(Database):
    def __init__(self, name: str, model_filename: str):
        models = PersistenceModels.from_filename(model_filename).get_models_names()
        super().__init__(name, models=models)

    def create(self):
        pass

    def delete(self):
        pass

    def clear_data(self):
        pass

    def get_model(self, model_name: str):
        model = self.models.get(model_name)
        if not model:
            raise IndexError(
                f'Model "{model_name}" is not available for "{self.name}" database'
            )
        return model

    def get_model_names(self) -> List[str]:
        return list(self.models.keys())



from functools import wraps
from inspect import signature

from meiga import Error, Failure
from meiga.on_failure_exception import OnFailureException

from petisco.base.domain.errors.unknown_error import UnknownError


def wrapper(execute_func, wrapped_class_name, config, mapper):
    @wraps(execute_func)
    def wrapped(*args, **kwargs):
        middlewares_classes = getattr(config, "middlewares", [])

        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")
        middlewares = [
            klass(wrapped_class_name, arguments) for klass in middlewares_classes
        ]

        for middleware in middlewares:
            middleware.before()

        try:
            result = execute_func(*args, **kwargs)
        except OnFailureException as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)

        mapped_result = mapper.map(result)
        for middleware in middlewares:
            if result:
                middleware.after(result)
        return mapped_result

    return wrapped



from meiga import Result

from petisco import __version__
from petisco.base.application.dependency_injection.injector import Injector
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.domain.errors.unknown_error import UnknownError


class NotifierMiddleware(Middleware):
    def __init__(self, wrapped_class_name, wrapped_class_input_arguments):
        super().__init__(wrapped_class_name, wrapped_class_input_arguments)
        self.notifier = Injector.get("notifier")

    def before(self):
        pass

    def after(self, result: Result):
        if result.is_failure:
            error = result.value
            if issubclass(error.__class__, UnknownError):
                notifier_exception_message = (
                    NotifierExceptionMessage.from_unknown_error(
                        error, title="Uncontrolled Exception"
                    )
                )
                notifier_exception_message.meta["petisco"] = __version__
                self.notifier.publish_exception(notifier_exception_message)



from typing import Dict
from abc import ABCMeta, abstractmethod

from petisco.legacy.notifier.domain.notifier_message import NotifierMessage


class INotifier:

    __metaclass__ = ABCMeta

    def __repr__(self):
        return "INotifier"

    @classmethod
    def info(cls) -> Dict:
        return {"name": cls.__name__}

    @abstractmethod
    def publish(self, notifier_message: NotifierMessage):
        raise NotImplementedError



# from petisco import INFO, event_from_redis_message
# from petisco.legacy.application.petisco import Petisco
#
#
# def redis_event_handler(message):
#     event = event_from_redis_message(message).unwrap_or_return()
#     logger = Petisco.get_instance()._logger
#     logger.log(INFO, event)



from abc import ABCMeta, abstractmethod

from petisco.legacy.tasks.config.config_tasks import ConfigTasks


class TaskExecutor:

    __metaclass__ = ABCMeta

    @abstractmethod
    def start(self, config_tasks: ConfigTasks):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError



from typing import Dict

from meiga import Result, Error

from petisco.legacy.application.petisco import Petisco
from petisco.legacy.controller.controller_handler import controller_handler
from petisco.legacy.controller.errors.http_error import HttpError
from petisco.legacy.modules.healthcheck.domain.persistence_error import PersistenceError
from petisco.legacy.modules.healthcheck.application.healthcheck_provider import (
    HealthcheckProvider,
)


class PersistenceHttpError(HttpError):
    def __init__(
        self,
        message: str = "Cannot connect with configured persistence. If you are not using persistence, just delete it from petisco.yml",
        code: int = 503,
    ):
        self.message = message
        self.code = code
        super(PersistenceHttpError, self).__init__(message, code)


def error_handler(result: Result) -> HttpError:
    domain_error = result.value
    http_error = HttpError()
    if isinstance(domain_error, PersistenceError):
        http_error = PersistenceHttpError()
    return http_error


@controller_handler(
    success_handler=lambda result: (result.value, 200), error_handler=error_handler
)
def healthcheck() -> Result[Dict, Error]:
    return HealthcheckProvider().execute(Petisco.get_instance())



from petisco.legacy.controller.errors.http_error import HttpError


class InternalHttpError(HttpError):
    def __init__(self, message: str = "Internal Error", code: int = 500, suffix=""):
        if suffix:
            message = message + " " + suffix

        super(InternalHttpError, self).__init__(message, code)



from contextlib import contextmanager
from typing import Callable

from petisco.legacy.persistence.elastic.elastic_operational_database_error import (
    ElasticOperationalDatabaseError,
)


def elastic_session_scope_provider(session) -> Callable:
    from elasticsearch import RequestError

    @contextmanager
    def session_scope():
        try:
            yield session
        except ConnectionRefusedError:
            raise ElasticOperationalDatabaseError
        except RequestError as e:  # noqa E722
            raise ElasticOperationalDatabaseError
        except:  # noqa E722
            raise ElasticOperationalDatabaseError

    return session_scope



from petisco.legacy.controller.errors.http_error import HttpError


class InvalidTokenHttpError(HttpError):
    def __init__(
        self,
        message: str = "Access token is missing or invalid.",
        code: int = 401,
        suffix="",
    ):
        if suffix:
            message = message + " " + suffix

        super(InvalidTokenHttpError, self).__init__(message, code)



from petisco.legacy.event.shared.domain.event import Event


class RabbitMqEventQueueNameFormatter:
    @staticmethod
    def format(event: Event, exchange_name: str = None) -> str:
        event_name = event.event_name.replace(".", "_")
        event_name = f"{event.event_version}.event.{event_name}"
        return f"{exchange_name}.{event_name}" if exchange_name else event_name

    @staticmethod
    def format_retry(event: Event, exchange_name: str = None) -> str:
        queue_name = RabbitMqEventQueueNameFormatter.format(event, exchange_name)
        return f"retry.{queue_name}"

    @staticmethod
    def format_dead_letter(event: Event, exchange_name: str = None) -> str:
        queue_name = RabbitMqEventQueueNameFormatter.format(event, exchange_name)
        return f"dead_letter.{queue_name}"



from abc import ABCMeta, abstractmethod
from typing import List, Dict

from petisco.legacy.notifier.domain.notifier_message import NotifierMessage


class ISlackNotifierMessageConverter:

    __metaclass__ = ABCMeta

    def __repr__(self):
        return "ISlackMessageConverter"

    @abstractmethod
    def convert(self, notifier_message: NotifierMessage) -> List[Dict]:
        """
        Should return a list of blocks. See: https://api.slack.com/reference/block-kit/blocks
        """
        raise NotImplementedError



from functools import partial
from typing import Dict

from meiga import Result, Error, isSuccess, Success, Failure

from petisco.legacy.domain.value_objects.client_id import ClientId
from petisco.legacy.domain.value_objects.name import Name
from petisco.legacy.domain.value_objects.user_id import UserId
from petisco.legacy.persistence.pymongo.pymongo_persistence_context import (
    get_mongo_collection,
)
from tests.end2end.flask_app.shared.domain.repositories.interface_user_repository import (
    IUserRepository,
)
from tests.end2end.flask_app.shared.domain.repositories.user_already_exist_error import (
    UserAlreadyExistError,
)
from tests.end2end.flask_app.shared.domain.repositories.user_not_found_error import (
    UserNotFoundError,
)
from tests.end2end.flask_app.user.domain.aggregate_roots.user import User


class MongoDBUserRepository(IUserRepository):
    @staticmethod
    def build():
        return MongoDBUserRepository()

    def __init__(self):
        self.collection_context = partial(get_mongo_collection, "user")

    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def save(self, user: User) -> Result[bool, Error]:
        result = self.exists(user.user_id)
        if result.is_success:
            return Failure(UserAlreadyExistError(user.user_id))

        with self.collection_context() as collection:
            if collection.insert_one(user.to_dict()):
                return isSuccess

    def retrieve(self, client_id: ClientId, user_id: UserId) -> Result[User, Error]:
        with self.collection_context() as collection:
            user_doc = collection.find_one(
                {"user_id": user_id.value, "client_id": client_id.value}
            )
            if user_doc:
                return Success(
                    User(
                        name=Name(user_doc.name),
                        client_id=ClientId(user_doc.client_id),
                        user_id=UserId(user_doc.user_id),
                    )
                )
            else:
                return Failure(UserNotFoundError(user_id))

    def exists(self, user_id: UserId) -> Result[bool, Error]:
        with self.collection_context() as collection:
            if collection.find_one({"user_id": user_id.value}):
                return isSuccess
            else:
                return Failure(UserNotFoundError(user_id))



import pytest

from petisco.legacy.domain.aggregate_roots.info_id import InfoId
from petisco.legacy.domain.value_objects.client_id import ClientId
from petisco.legacy.domain.value_objects.name import Name
from petisco.legacy.domain.value_objects.user_id import UserId
from petisco.legacy.domain.value_objects.correlation_id import CorrelationId


@pytest.fixture
def given_any_client_id() -> ClientId:
    return ClientId("petisco-client")


@pytest.fixture
def given_any_user_id() -> UserId:
    return UserId("9f74b5c0-2196-4a27-a03f-cb92d66e2bbd")


@pytest.fixture
def given_any_correlation_id() -> CorrelationId:
    return CorrelationId("fde825e8-8ac2-4ca1-950a-5c74c8f6b66d")


@pytest.fixture
def given_any_info_id(
    given_any_client_id, given_any_user_id, given_any_correlation_id
) -> InfoId:
    return InfoId(given_any_client_id, given_any_user_id, given_any_correlation_id)


@pytest.fixture
def given_info_id_only_correlation_id(given_any_correlation_id) -> InfoId:
    return InfoId(correlation_id=given_any_correlation_id)


@pytest.fixture
def given_any_name() -> Name:
    return Name("Petisco")



from meiga import Error

from petisco import Uuid
from petisco.base.domain.ids.client_id import ClientId
from petisco.base.domain.ids.user_id import UserId


class ClientNotFoundError(Error):
    def __init__(self, client_id: ClientId):
        self.message = f"{self.__class__.__name__} [client_id: {client_id.value}]"


class ClientAlreadyExistError(Error):
    def __init__(self, client_id: ClientId):
        self.message = f"{self.__class__.__name__} [client_id: {client_id.value}]"


class UserNotFoundError(Error):
    def __init__(self, user_id: UserId):
        self.message = f"{self.__class__.__name__} [user_id: {user_id.value}]"


class UserAlreadyExistError(Error):
    def __init__(self, user_id: UserId):
        self.message = f"{self.__class__.__name__} [user_id: {user_id.value}]"


class EntityAlreadyExistError(Error):
    def __init__(
        self,
        repository_name: str = None,
        table_name: str = None,
        entity_id: Uuid = None,
    ):
        entity_id_str = f" ({entity_id.value})" if entity_id else ""
        repository_str = f" (repository: {repository_name})" if repository_name else ""
        table_str = f" (table: {table_name})" if repository_name else ""
        self.message = f"Entity{entity_id_str} already exist{repository_str}{table_str}"


class EntityNotFoundError(Error):
    def __init__(self, repository_name: str = None, entity_id: Uuid = None):
        entity_id_str = f" ({entity_id.value})" if entity_id else ""
        repository_str = f" (repository: {repository_name})" if repository_name else ""
        self.message = f"Entity{entity_id_str} not found{repository_str}"


class EntitiesNotFoundError(Error):
    def __init__(self, repository_name: str = None):
        repository_str = f" (repository: {repository_name})" if repository_name else ""
        self.message = f"Entities not found{repository_str}]"



from typing import Dict

from petisco.legacy.event.shared.domain.event import Event
from petisco.legacy.event.legacy.publisher.domain.interface_event_publisher import (
    IEventPublisher,
)


class NotImplementedEventPublisher(IEventPublisher):
    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def publish(self, event: Event):
        pass

    def stop(self):
        pass



from typing import List, Callable

from petisco.base.domain.persistence.interface_database import Database
from petisco.extra.elastic.elastic_connection import ElasticConnection
from petisco.extra.elastic.elastic_session_scope_provider import (
    elastic_session_scope_provider,
)


class ElasticDatabase(Database):
    @staticmethod
    def local_connection_checker():
        return ElasticDatabase(name="test", connection=ElasticConnection.create_local())

    def __init__(self, name: str, connection: ElasticConnection):
        if not connection or not isinstance(connection, ElasticConnection):
            raise ConnectionError(
                "ElasticDatabase needs a valid ElasticConnection connection"
            )
        self.connection = connection
        super().__init__(name)

    def create(self):
        from elasticsearch import Elasticsearch

        self.session = Elasticsearch(
            self.connection.to_elastic_format(), http_auth=self.connection.http_auth
        )

    def delete(self):
        pass

    def clear_data(self):
        pass

    def is_available(self):
        try:
            return self.get_session().ping()
        except Exception:  # noqa E722
            return False

    def get_base(self):
        return None

    def get_model(self, model_name: str):
        return None

    def get_model_names(self) -> List[str]:
        return None

    def get_session(self):
        return self.session

    def get_session_scope(self) -> Callable:
        return elastic_session_scope_provider(self.get_session())



from petisco.legacy.domain.value_objects.correlation_id import CorrelationId


def flask_correlation_id_provider(func_name: str = "") -> CorrelationId:
    try:
        import connexion

        correlation_id_str = connexion.request.headers.get("X-Correlation-Id")
        if correlation_id_str:
            correlation_id = CorrelationId(correlation_id_str)
        else:
            correlation_id = CorrelationId.generate(func_name)
    except (RuntimeError, ImportError):
        correlation_id = CorrelationId.generate(func_name)
    return correlation_id



from typing import List
from abc import abstractmethod

from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.base.misc.interface import Interface


class MessageConfigurer(Interface):
    @abstractmethod
    def configure_subscribers(self, subscribers: List[MessageSubscriber]):
        raise NotImplementedError

    @abstractmethod
    def clear(self):
        raise NotImplementedError



import hashlib

from meiga import Error

from petisco.legacy.domain.value_objects.value_object import ValueObject


class SignatureAlgorithmIsNotAvailableError(Error):
    def __init__(self, message):
        self.message = message


class SignatureAlgorithm(ValueObject):
    @staticmethod
    def sha256():
        return SignatureAlgorithm("sha256")

    @staticmethod
    def sha512():
        return SignatureAlgorithm("sha512")

    def __init__(self, value: str):
        self.value = value
        self.available_hash_algoritmhs = {
            "sha256": hashlib.sha256,
            "sha512": hashlib.sha512,
        }
        super(SignatureAlgorithm, self).__init__()

    def guard(self):
        if self.value not in self.available_hash_algoritmhs.keys():
            raise SignatureAlgorithmIsNotAvailableError(self.value)

    def get_algorithm(self):
        return self.available_hash_algoritmhs.get(self.value)



from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus


class NotImplementedDomainEventBus(DomainEventBus):
    def publish(self, domain_event: DomainEvent):
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        _ = domain_event.update_meta(meta)

    def retry_publish_only_on_store_queue(self, domain_event: DomainEvent):
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        _ = domain_event.update_meta(meta)

    def close(self):
        pass



from typing import Callable, Dict

from dataclasses import dataclass

from petisco.legacy.application.config.get_funtion_from_string import (
    get_function_from_string,
)
from petisco.legacy.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


def get_param(kdict: Dict, key: str):
    value = kdict.get(key)
    if not value:
        raise TypeError(f"ConfigEventSubscriber: {key} is a required parameter")
    return value


@dataclass
class ConfigEventSubscriber:
    organization: str
    service: str
    topic: str
    dead_letter: bool = False
    handler_key: str = None
    handler: Callable = None
    kdict: Dict = None

    def __repr__(self):
        return f"EventSubscriberConfig: {self.organization} | {self.service} | {self.topic} | {self.dead_letter} -> {self.handler}"

    @staticmethod
    def from_dict(kdict: Dict):
        organization = get_param(kdict, "organization")
        service = get_param(kdict, "service")
        handler_key = get_param(kdict, "handler")
        topic = get_param(kdict, "topic")
        dead_letter = kdict.get("dead_letter", False)

        return ConfigEventSubscriber(
            organization=organization,
            service=service,
            handler_key=handler_key,
            topic=topic,
            dead_letter=dead_letter,
            kdict=kdict,
        )

    def get_handler(self):
        # Use this function to avoid circular dependency accessing Petisco objects before its configuration

        if self.handler_key:
            handler = (
                get_function_from_string(self.handler_key)
                .handle(
                    on_failure=raise_petisco_config_exception,
                    failure_args=(self.kdict, "event:subscriber:subscriber"),
                )
                .unwrap()
            )
        else:
            handler = self.handler

        return handler



from typing import Optional

from pydantic.main import BaseModel

from petisco.base.domain.ids.client_id import ClientId
from petisco.base.domain.ids.correlation_id import CorrelationId
from petisco.base.domain.ids.user_id import UserId


class InfoId(BaseModel):
    client_id: Optional[ClientId] = None
    user_id: Optional[UserId] = None
    correlation_id: Optional[CorrelationId] = None
    ip: Optional[str] = None  # TODO Review IPv4Address

    def __init__(
        self,
        client_id: Optional[ClientId] = None,
        user_id: Optional[UserId] = None,
        correlation_id: Optional[CorrelationId] = None,
        ip: Optional[str] = None,
        **data
    ) -> None:
        super().__init__(
            client_id=client_id,
            user_id=user_id,
            correlation_id=correlation_id,
            ip=ip,
            **data
        )

    @staticmethod
    def from_headers(headers: dict):
        if headers:
            info_id = InfoId(
                client_id=headers.get("X-Onboarding-Clientid"),
                user_id=headers.get("X-Onboarding-Userid"),
                correlation_id=headers.get("X-Correlation-Id"),
                ip=headers.get("X-Forwarded-For"),
            )
        else:
            info_id = InfoId()
        return info_id

    def get_http_headers(self):
        headers = {}

        if self.client_id:
            headers["X-Onboarding-Clientid"] = self.client_id.value
        if self.user_id:
            headers["X-Onboarding-Userid"] = self.user_id.value
        if self.correlation_id:
            headers["X-Correlation-Id"] = self.correlation_id.value
        if self.ip:
            headers["X-Forwarded-For"] = str(self.ip)

        return headers if len(headers) > 0 else None



try:
    from .domain import *  # noqa F403
    from .application import *  # noqa F403
    from .persistence import *  # noqa F403
    from .domain import *  # noqa F403
    from .client import *  # noqa F403
    from .testing_decorators import *  # noqa F403
except (RuntimeError, ImportError):
    print(
        "Impossible to import fixtures. fixtures petisco extra is required, install it with `pip install petisco[fixtures]`"
    )



from contextlib import contextmanager

from pymongo.errors import InvalidName, PyMongoError

from petisco.legacy.persistence.pymongo.pymongo_errors import (
    PyMongoOperationalDatabaseError,
    PyMongoInvalidCollectionNameError,
    PyMongoInvalidDatabaseNameError,
)
from petisco.legacy.persistence.pymongo.pymongo_persistence import PyMongoPersistence


@contextmanager
def get_mongo_collection(collection: str):
    client = PyMongoPersistence.get_instance().client
    database = PyMongoPersistence.get_instance().database
    try:
        db = client.get_database(database)
    except InvalidName as e:
        print(e)
        raise PyMongoInvalidDatabaseNameError(e)
    try:
        yield db.get_collection(collection)
    except InvalidName as e:
        print(e)
        raise PyMongoInvalidCollectionNameError(e)
    except PyMongoError as e:
        print(e)
        raise PyMongoOperationalDatabaseError(e)



import os

from setuptools import find_packages, setup

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
PACKAGE_NAME = "petisco"
VERSION = open("petisco/VERSION", "r").read().rstrip()

# The text of the README file
with open(os.path.join(CURRENT_DIR, "README.md")) as fid:
    README = fid.read()

with open("requirements/requirements.txt") as f:
    required = f.read().splitlines()

legacy_packages = [
    "petisco/legacy",
    "petisco/legacy/controller",
    "petisco/legacy/controller/errors",
    "petisco/legacy/domain",
    "petisco/legacy/domain/value_objects",
    "petisco/legacy/domain/aggregate_roots",
    "petisco/legacy/domain/errors",
    "petisco/legacy/notifier",
    "petisco/legacy/notifier/config",
    "petisco/legacy/notifier/infrastructure",
    "petisco/legacy/notifier/infrastructure/slack",
    "petisco/legacy/notifier/domain",
    "petisco/legacy/tasks",
    "petisco/legacy/tasks/config",
    "petisco/legacy/tasks/domain",
    "petisco/legacy/tasks/infrastructure",
    "petisco/legacy/fixtures",
    "petisco/legacy/frameworks",
    "petisco/legacy/frameworks/flask",
    "petisco/legacy/frameworks/flask/application",
    "petisco/legacy/http",
    "petisco/legacy/logger",
    "petisco/legacy/persistence",
    "petisco/legacy/persistence/sqlalchemy",
    "petisco/legacy/persistence/pymongo",
    "petisco/legacy/persistence/sql",
    "petisco/legacy/persistence/sql/mysql",
    "petisco/legacy/persistence/sql/sqlite",
    "petisco/legacy/persistence/elastic",
    "petisco/legacy/security/token_decoder",
    "petisco/legacy/security/token_manager",
    "petisco/legacy/tools",
    "petisco/legacy/use_case",
    "petisco/legacy/modules",
    "petisco/legacy/modules/environment",
    "petisco/legacy/modules/environment/application",
    "petisco/legacy/modules/environment/domain",
    "petisco/legacy/modules/healthcheck",
    "petisco/legacy/modules/healthcheck/application",
    "petisco/legacy/modules/healthcheck/domain",
]

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description="Petisco is a framework for helping Python developers to build clean Applications",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=["DDD", "Use Case", "Clean Architecture", "REST", "Applications"],
    url="https://github.com/alice-biometrics/petisco",
    author="Alice Biometrics",
    author_email="support@alicebiometrics.com",
    license="MIT",
    install_requires=required,
    entry_points={
        "console_scripts": ["petisco = petisco.legacy.application.cli.petisco:main"]
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages() + legacy_packages,
    include_package_data=True,
    zip_safe=False,
    extras_require={
        "flask": [
            "connexion==2.6.0",
            "swagger-ui-bundle>=0.0.2",
            "Flask-Cors==3.0.7",
            "Flask-Bcrypt==0.7.1",
        ],
        "sqlalchemy": [
            "sqlalchemy==1.4.20",
            "sqlalchemy_utils==0.37.8",
            "PyMySQL==1.0.2",
        ],
        "redis": ["redis >= 3.3.11", "fakeredis>=1.0.5"],
        "rabbitmq": ["pika==1.2.0"],
        "gunicorn": ["gunicorn", "json-logging-py==0.2"],
        "fixtures": ["pytest"],
        "slack": ["slackclient"],
        "pymongo": ["pymongo==3.11.0"],
        "elastic": ["elasticsearch<8.0.0,>=7.13.1"],
        "fastapi": ["fastapi==0.65.2"],
    },
)



from typing import Dict

from pydantic.main import BaseModel


class ConsumerDerivedAction(BaseModel):
    action: str = None
    exchange_name: str = None
    routing_key: str = None
    headers: Dict = None
    # def __init__(
    #     self,
    #     action: str = None,
    #     exchange_name: str = None,
    #     routing_key: str = None,
    #     headers: Dict = None,
    # ):
    #     self.action = action
    #     self.exchange_name = exchange_name
    #     self.routing_key = routing_key
    #     self.headers = headers
    #
    # def to_dict(self):
    #     if not self.action:
    #         return {}
    #     else:
    #         return {
    #             "action": self.action,
    #             "exchange_name": self.exchange_name,
    #             "routing_key": self.routing_key,
    #             "headers": self.headers,
    #         }



import pprint

import six
import typing

from petisco.legacy.frameworks.flask.application import util

T = typing.TypeVar("T")


class Model(object):
    # openapiTypes: The key is attribute name and the
    # value is attribute type.
    openapi_types = {}

    # attributeMap: The key is attribute name and the
    # value is json key in definition.
    attribute_map = {}

    @classmethod
    def from_dict(cls: typing.Type[T], dikt) -> T:
        """Returns the dict as a model"""
        return util.deserialize_model(dikt, cls)

    def to_dict(self):
        """Returns the model properties as a dict

        :rtype: dict
        """
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model

        :rtype: str
        """
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



from petisco.base.domain.errors.domain_error import DomainError


class AlreadyExists(DomainError):
    def get_specify_detail(self) -> str:
        return "Already Exists"


class ClientAlreadyExists(DomainError):
    def get_specify_detail(self) -> str:
        return "Client Already Exists"


class UserAlreadyExists(DomainError):
    def get_specify_detail(self) -> str:
        return "User Already Exists"



from typing import Dict
from abc import ABCMeta, abstractmethod

from petisco.legacy.event.legacy.subscriber.domain.config_event_subscriber import (
    ConfigEventSubscriber,
)


class IEventSubscriber:

    __metaclass__ = ABCMeta

    def __init__(self, subscribers: Dict[str, ConfigEventSubscriber]):
        self.subscribers = subscribers

    def __repr__(self):
        return f"IEventPublisher: [subscribers: {self.subscribers}]"

    @abstractmethod
    def info(self) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def start(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError



from petisco.legacy import Event


class UserCreated(Event):
    name: str

    def __init__(self, name: str, version: int = 1):
        self.name = name
        self.event_version = version
        super().__init__()



import copy

from petisco.base.misc.interface import Interface


class BasePattern(Interface):
    def with_info_id(self, info_id):
        base = copy.copy(self)
        base._set_info_id(info_id)
        return base

    def _set_info_id(self, info_id):
        if info_id is not None:
            self.info_id = info_id

    def get_info_id(self):
        if not hasattr(self, "info_id"):
            name = self.info().get("name")
            raise AttributeError(
                f"{name} needs info_id configuration. Please, use {name}.info_id() to get a valid object with info_id"
            )
        return self.info_id



import os

import pytest

from petisco.legacy.application.petisco import Petisco


@pytest.fixture
def petisco_sql_database(request):
    persistence_source_marker = request.node.get_closest_marker("persistence_source")
    if persistence_source_marker is None:
        persistence_source_name = "petisco"
    else:
        persistence_source_name = persistence_source_marker.args[0]

    try:
        petisco = Petisco.get_instance()
        persistence_source_config = petisco.legacy.persistence_sources[
            persistence_source_name
        ]
    except ImportError as e:
        raise ImportError(
            f"{e}\n petisco_sql_database needs a preconfigured Petisco to have available Model definitions"
        )
    except KeyError as e:
        raise KeyError(
            f"{e}\n petisco_sql_database needs a preconfigured Petisco with persistence. Required persistence ({persistence_source_name}) is not available"
        )

    if not persistence_source_config or not persistence_source_config["configured"]:
        yield
    else:
        from sqlalchemy import create_engine
        from petisco.legacy.persistence.sqlalchemy.sqlalchemy_persistence import (
            SqlAlchemyPersistence,
        )

        sql_database = os.environ.get("SQL_DATABASE")
        if not sql_database:
            raise RuntimeError(
                "Please add required SQL_DATABASE env variable (e.g pytest.ini, pytest.end2end.ini)"
            )

        Base = SqlAlchemyPersistence.get_instance().sources[persistence_source_name][
            "base"
        ]
        Session = SqlAlchemyPersistence.get_instance().sources[persistence_source_name][
            "session"
        ]
        connection = f"sqlite:///{sql_database}"
        engine = create_engine(connection)
        Base.metadata.create_all(engine)

        yield

        session = Session()
        session.rollback()
        session.close()
        Base.metadata.drop_all(bind=engine)
        os.remove(sql_database)



from abc import ABCMeta, abstractmethod

from meiga import Result, NotImplementedMethodError

from petisco.legacy.application.pattern_base import PatternBase


class IAppService(PatternBase):

    __metaclass__ = ABCMeta

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError


# Deprecated
class IService(PatternBase):

    __metaclass__ = ABCMeta

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError



import importlib
import os
import sys
import traceback
from typing import Dict, Any

import yaml
from yaml.parser import ParserError, ScannerError


class PersistenceModels:
    def __init__(self, models: Dict[str, Any]):
        self.models = models
        self.imported_models = {}

    @staticmethod
    def from_filename(filename: str):
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"PersistenceModels ({filename} not found)")
        try:
            with open(filename) as file:
                petisco_dict = yaml.load(file, Loader=yaml.FullLoader)
                models = petisco_dict.get("models")
                return PersistenceModels(models)
        except (ParserError, ScannerError) as e:
            message = f"Error loading {filename} file: {repr(e.__class__)} {e} | {traceback.format_exc()}"
            raise RuntimeError(message)

    def get_models_names(self):
        return self.models

    def import_models(self):
        self.imported_models = self.import_database_models()

        # self.imported_models = {}
        # for name, model_string in self.models.items():
        #     mod_name, model_name = model_string.rsplit(".", 1)
        #     mod = importlib.import_module(mod_name)
        #     self.imported_models[name] = getattr(mod, model_name)

    def import_database_models(self):
        def _is_class_in_sqlalchemy_tables(class_model_name: str):
            # We need this to filter deletion in case of Model dependencies.
            # For instance, ClientConfigModel imports ClientModel on class model definition
            # It throws the following error:
            #  sqlalchemy.exc.InvalidRequestError: Table 'ClientConfig' is already defined for this MetaData instance.
            #  Specify 'extend_existing=True' to redefine options and columns on an existing Table object.
            # As a workaround, we only delete the module if it not exist in SQLAlchemy tables.

            from petisco.base.domain.persistence.persistence import Persistence

            is_in_tables = False
            for database in Persistence.get_databases():
                if not hasattr(database, "get_base"):
                    continue
                Base = database.get_base()
                if not Base:
                    continue
                is_in_tables = class_model_name.replace("Model", "") in list(
                    Base.metadata.tables.keys()
                )
            return is_in_tables

        def _delete_module_if_already_imported(module_name: str, class_model_name: str):
            # Important. SqlAlchemy needs models to be imported
            # Base = Persistence.get_base("petisco")
            # If module is already imported and Base has not tables, Tables won't be imported.
            # This, usually happens when start the application (e.g end2end tests)
            # e.g len(Base.metadata.tables) == 0
            # With this function we ensure that model can be imported to help SqlAlchemy loading models if is not imported yet
            if module_name in sys.modules and not _is_class_in_sqlalchemy_tables(
                class_model_name
            ):
                del sys.modules[module_name]

        def _import_database_models_func():
            imported_models = {}
            for name, model_string in self.models.items():
                module_name, class_model_name = model_string.rsplit(".", 1)
                _delete_module_if_already_imported(module_name, class_model_name)
                module_name = importlib.import_module(module_name, class_model_name)
                imported_models[name] = getattr(module_name, class_model_name)
            return imported_models

        return _import_database_models_func()

    def get_imported_models(self):
        return self.imported_models



from typing import Callable, Optional

from meiga import Result
from pika import BasicProperties
from pika.spec import Basic

from petisco.base.domain.message.consumer_derived_action import ConsumerDerivedAction
from petisco.legacy.logger.interface_logger import ILogger, DEBUG
from petisco.legacy.logger.log_message import LogMessage
from petisco.legacy.logger.not_implemented_logger import NotImplementedLogger


class RabbitMqMessageConsumerLogger:
    def __init__(self, logger: Optional[ILogger] = NotImplementedLogger()):
        self.logger = logger

    def _get_base_message(self, handler: Callable):
        return LogMessage(
            layer="rabbitmq_message_consumer", operation=f"{handler.__name__}"
        )

    def _get_event_handler_name(self, handler: Callable):
        handler_name = getattr(handler, "__name__", repr(handler))
        handler_module = getattr(handler, "__module__") + "."
        return f"{handler_module}{handler_name}"

    def log_nack_simulation(
        self,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        handler: Callable,
    ):
        self._log_simulation(method, properties, body, handler, "nack simulated")

    def log_failure_simulation(
        self,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        handler: Callable,
    ):
        self._log_simulation(method, properties, body, handler, "failure simulated")

    def _log_simulation(
        self,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        handler: Callable,
        chaos_action: str,
    ):
        log_message = self._get_base_message(handler)
        event_handler_name = self._get_event_handler_name(handler)
        message = {
            "body": body,
            "properties": properties,
            "method": method,
            "event_handler": event_handler_name,
            "chaos_action": chaos_action,
        }
        self.logger.log(DEBUG, log_message.set_message(message))

    def log_parser_error(
        self,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        handler: Callable,
        exception: Exception,
    ):
        log_message = self._get_base_message(handler)
        event_handler_name = self._get_event_handler_name(handler)
        message = {
            "body": body,
            "properties": properties,
            "method": method,
            "event_handler": event_handler_name,
            "exception": str(exception),
        }
        self.logger.log(DEBUG, log_message.set_message(message))

    def log(
        self,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        handler: Callable,
        log_activity: str = None,
        result: Result = None,
        derived_action: ConsumerDerivedAction() = None,
    ):
        log_message = self._get_base_message(handler)
        event_handler_name = self._get_event_handler_name(handler)

        message = {
            "body": body,
            "properties": properties,
            "method": method,
            "event_handler": event_handler_name,
        }
        if log_activity:
            message["log_activity"] = log_activity
        if result:
            message["result"] = result
        if derived_action:
            message["derived_action"] = derived_action.dict()

        self.logger.log(DEBUG, log_message.set_message(message))



from petisco.legacy.domain.value_objects.value_object import ValueObjectError


class ExceedLengthLimitValueObjectError(ValueObjectError):
    def __init__(self, message):
        self.message = message


class NotReachMinimumValueObjectError(ValueObjectError):
    def __init__(self, message):
        self.message = message


class NotHasSpecificLengthValueObjectError(ValueObjectError):
    def __init__(self, message):
        self.message = message



from petisco.base.domain.errors.domain_error import DomainError


class NotFound(DomainError):
    def get_specify_detail(self) -> str:
        return "Not Found"


class ClientNotFound(DomainError):
    def get_specify_detail(self) -> str:
        return "Client Not Found"


class UserNotFound(DomainError):
    def get_specify_detail(self) -> str:
        return "User Not Found"
